# -*- coding: utf-8 -*-

from typing import Optional
from ..utils.encoding import encode_tx_hash


class TxHash:
    def __init__(self, th: Optional[bytes]) -> None:
        self.__tx_hash = th

    def __str__(self) -> str:
        hash_str = encode_tx_hash(self.__tx_hash)
        return '' if hash_str is None else hash_str

    def __bytes__(self) -> bytes:
        return self.__tx_hash if self.__tx_hash else b''



import aergo.herapy as herapy


def test_aergo_enterprise_contract() -> None:
    SC_ADDRESS = herapy.GovernanceTxAddress.ENTERPRISE
    """
    const SetConf = "setConf"
    const AppendConf = "appendConf"
    const RemoveConf = "removeConf"
    const AppendAdmin = "appendAdmin"
    const RemoveAdmin = "removeAdmin"
    const EnableConf = "enableConf"
    const DisableConf = "disableConf"
    const ChangeCluster = "changeCluster"
    """
    aergo = herapy.Aergo()

    print("------ Connect AERGO -----------")
    aergo.connect('localhost:7848')

    print("------ Set Admin Account -----------")
    sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
    sender_account = aergo.new_account(private_key=sender_private_key)
    print("  > Admin Address: {}".format(sender_account.address))
    assert sender_account.address
    print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))

    print("------ Call SC: appendAdmin -----------")
    tx, result = aergo.call_sc(
        SC_ADDRESS, "appendAdmin", args=[str(sender_account.address)])
    print(result)

    print("-------Wait for tx result--------")
    result = aergo.wait_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.SUCCESS, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    print("------ Check result of Call SC -----------")
    print("  > TX: {}".format(tx.tx_hash))
    result = aergo.get_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.SUCCESS, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    config = aergo.get_enterprise_config('ADMINS')
    print(config.values)
    assert config.values == [str(sender_account.address)]

    print("------ Call SC: removeAdmin -----------")
    tx, result = aergo.call_sc(
        SC_ADDRESS, "removeAdmin", args=[str(sender_account.address)])

    print("-------Wait for tx result--------")
    result = aergo.wait_tx_result(tx.tx_hash)
    print(result)
    assert result.status == herapy.TxResultStatus.SUCCESS, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    print("------ Check result of Call SC -----------")
    print("  > TX: {}".format(tx.tx_hash))
    result = aergo.get_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.SUCCESS, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)
    config = aergo.get_enterprise_config('ADMINS')
    print(config.values)
    assert config.values == []



# -*- coding: utf-8 -*-

__all__ = ["exception"]

from .exception import AergoException, CommunicationException
from .conversion_exception import ConversionException
from .general_exception import GeneralException



import aergo.herapy as herapy


def test_sc(aergo) -> None:
    print("------ Payload -----------")
    """
        -- Define global variables.
        state.var {
            a = state.value(),
        }

        function constructor(name)
            a:set(name)
        end

        function set_name(name)
            a:set(name)
        end

        function get_name()
            return a:get()
        end

        function test_array(array)
            a:set(array[1])
        end

        function say_hello()
            a:set("hello")
        end


        abi.register(set_name, get_name, test_array, say_hello)
    """
    payload_str = "2bTAKwCcCqs1n91vFF5YnBBcaTLNsxhqzGgM8vcVAG3HqQLp1uZZSwvmn" \
        "j8CEEm9GTvogDf3r6ioqHjbfBisjRzwUJxr6msFuf4vVuzbWEXtbShu43ZpLbJdT1NE" \
        "CStCEDGFPRsn3pc7jQ1FumPGB9GwGY7Y7wNN1t4kLMQPNxJCmDFAwLA6BAyCGpzW2g5" \
        "rtHuWodTMTposF4EBRxbvsXYWvrDYLTK7JGbDsq4neBzXVtmJjereb7i6fX8Fug1LJ9" \
        "GgYskTKdqcBjugxEoipjsAALtnr7mcUPoPFygbQzXbHsW13UwjhBxKsEvWYZm1hC8Qx" \
        "CeiRbHdx8VY5ES2ugR8Dt4uhiboAbaLofpnGRcPofgFGXiwDfESQgNZBjHP1DeEzh7g" \
        "Gh6fzABcw3LSjQLC9RETFWyUnWY3dU4iqH3PsVNhBFgHCUepBBphvwPT6UoictgsWzs" \
        "nFdf1peiDsfXFE9uSTKWDCMhqgaETSLZdn8QiGBwMvZ1pEkP6xmh9389mFXCZ4ERNGU" \
        "Gkk6xNfN7xdZhRCeXxZupgtNDREX5PrHPveXEMbzThAMV7Rqzy3MazDaCFxjkbUVgQj" \
        "WsHYhwEBRMvYSd2p621nZzcb9GFbh9tpXhwoLYpLp89Qhi9oTy61AD2GVFgQmdWF9ne" \
        "D1GvauVkCNFzWmEUkdSU3F6yBRwBm2cs8oJEieFu1zETLsMyXydTN5WqHFuGs2PaV1X" \
        "RwPVRLeYiwo2xmgJMssHaGtR8pmb8TrQpTJbJdgreCVn8GiGzYizvEcRz9m8aFdompJ" \
        "2m2QR7TDLCCjZx2UqRJdcQZmyxEAtzjnCCgD7Gwj95ZHkFxzoEEHTLYbG6uaneRrQWn" \
        "zabAaRe2392qJnEkiNug9gkdJfsRFapbVxupH9zkoAmFP5xoWeGrkknUw61E4tX"
    payload = herapy.utils.decode_contract_code(payload_str)
    print(''.join('{:d} '.format(x) for x in payload))
    assert herapy.utils.encode_contract_code(payload) == payload_str

    byte_code_len = int.from_bytes(payload[:4], byteorder='little')
    print("payload: byte code length = ", byte_code_len)

    print("------ Set Sender Account -----------")
    sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
    sender_account = aergo.new_account(private_key=sender_private_key)
    print("  > Sender Address: {}".format(sender_account.address))
    print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))

    print("------ Set Receiver Account -----------")
    receiver_address = "AmNHbk46L5ZaFH942mxDrunhUb34S8xRd7ygNnqaW5nqJDt5ugKD"
    print("  > Receiver Address: {}".format(receiver_address))

    print("------ Deploy SC -----------")
    tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234)
    print("  > TX: {}".format(tx.tx_hash))
    print("{}".format(herapy.utils.convert_tx_to_json(tx)))
    assert result.status == herapy.CommitStatus.TX_OK, \
        "    > ERROR[{0}]: {1}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
    print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

    aergo.wait_tx_result(tx.tx_hash)
    aergo.get_account()

    print("------ Check deployment of SC -----------")
    print("  > TX: {}".format(tx.tx_hash))
    result = aergo.get_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.CREATED, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    sc_address = result.contract_address
    print("  > SC Address: {}".format(sc_address))

    print("------ Fail SC -----------")
    tx, result = aergo.call_sc(sc_address, "set_none")
    aergo.wait_tx_result(tx.tx_hash)
    result = aergo.get_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.ERROR, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    print("------ Query SC -----------")
    result = aergo.query_sc(sc_address, "get_name")
    assert result == b'1234'

    print("------ Call SC -----------")
    tx, result = aergo.call_sc(sc_address, "test_array", args=[["a", "b"]])

    print("-------Wait for tx result--------")
    result = aergo.wait_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.SUCCESS, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    print("------ Check result of Call SC -----------")
    print("  > TX: {}".format(tx.tx_hash))
    result = aergo.get_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.SUCCESS, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    print("------ Query SC -----------")
    result = aergo.query_sc(sc_address, "get_name")
    assert result == b'"a"'

    print("------- Get smart contract abi --------")
    abi = aergo.get_abi(sc_address)
    assert abi.functions[0] == {'name': 'say_hello'}



# -*- coding: utf-8 -*-

from typing import Union
from ..utils.encoding import encode_block_hash, decode_block_hash


class BlockHash:
    def __init__(self, bh: Union[str, bytes]) -> None:
        if isinstance(bh, str):
            hash_bytes = decode_block_hash(bh)
            assert hash_bytes  # for mypy Optional[bytes] -> bytes
            bh = hash_bytes
        self.__block_hash = bh

    @property
    def value(self) -> bytes:
        return self.__block_hash

    def __bytes__(self) -> bytes:
        return self.__block_hash

    def __str__(self) -> str:
        hash_str = encode_block_hash(self.__block_hash)
        assert hash_str
        return hash_str



import aergo.herapy as herapy


def test_tx(aergo) -> None:
    print("------ Set Sender Account -----------")
    sender_private_key = \
        "eHoEcHnaxpGpgzknXjuwon8VFVrLkKHC4FckGuGkQ8depiDDfyUAWC3L"
    sender_account = aergo.new_account(private_key=sender_private_key)
    print("  > Sender Address: {}".format(sender_account.address))
    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))

    print("------ Set Receiver Account -----------")
    receiver_address = "AmMPQqRJ4pd8bS9xdkw5pjExtLjaUXaYGB5kagzHu4A9ckKgnBV2"
    print("  > Receiver Address: {}".format(receiver_address))
    receiver_account = aergo.get_account(address=receiver_address)
    print("    > account state of Sender")
    print("      - balance        = {}".format(receiver_account.balance))
    print("      - nonce          = {}".format(receiver_account.nonce))
    print("      - code hash      = {}".format(receiver_account.code_hash))
    print("      - storage root   = {}".format(receiver_account.storage_root))

    print("------ Simple Send Tx -----------")
    simple_tx, result = aergo.send_payload(to_address=receiver_address,
                                           amount=10, payload=None,
                                           retry_nonce=3)
    print("  > simple TX[{}]".format(simple_tx.calculate_hash()))
    print("{}".format(str(simple_tx)))
    print("  > result: ", result)
    assert result.status == herapy.CommitStatus.TX_OK, \
        "    > ERROR[{0}]: {1}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))

    aergo.wait_tx_result(simple_tx.tx_hash)

    print("------ Check TX in Aergo -----------", flush=True)
    tx = aergo.get_tx(simple_tx.tx_hash)
    print(" TX from Aergo: ", str(tx))
    tx = tx.block.get_tx(tx.index_in_block)
    print(" TX in Block: ", str(tx))

    print("------ Check TX status -----------")
    print("  > TX: {}".format(str(simple_tx.tx_hash)))
    tx_result = aergo.get_tx_result(simple_tx.tx_hash)
    print("      result: ", tx_result)
    assert tx_result.status == herapy.TxResultStatus.SUCCESS, \
        "  > ERROR[{0}]:{1}: {2}".format(
            tx_result.contract_address, tx_result.status, tx_result.detail)

    print("------ Get Sender Account Info -----------")
    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))

    print("------ Create Tx -----------")
    tx = herapy.Transaction(from_address=bytes(sender_account.address),
                            nonce=sender_account.nonce + 1,
                            amount=10)
    tx.to_address = herapy.utils.decode_address(receiver_address)
    tx.chain_id = aergo.chain_id
    print("  > unsigned TX Hash: {}".format(tx.tx_hash))
    print("  > unsigned TX : {}"
          .format(herapy.utils.convert_tx_to_formatted_json(tx)))

    print("------ Sign Tx -----------")
    tx.sign = sender_account.sign_msg_hash(
        tx.calculate_hash(including_sign=False))
    print("  > TX Signature: {}".format(tx.sign_str))
    print("  > TX Hash: {}".format(tx.tx_hash))
    print("  > TX : {}".format(herapy.utils.convert_tx_to_formatted_json(tx)))

    print("------ Send Tx -----------")
    tx, result = aergo.send_tx(tx)
    print("{}".format(herapy.utils.convert_tx_to_formatted_json(tx)))
    assert result.status == herapy.CommitStatus.TX_OK, \
        "    > ERROR[{0}]: {1}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))

    aergo.wait_tx_result(tx.tx_hash)

    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))



def test_node_account(aergo) -> None:
    aergo.new_account(skip_state=True)

    print("------ Get Accounts without State in Node -----------")
    accounts = aergo.get_node_accounts(skip_state=True)
    for i, account in enumerate(accounts):
        print("  > account state : {}".format(account))
        print("    - balance")
        print("        + aergo              = {}"
              .format(account.balance.aergo))
        print("        + gaer               = {}".format(account.balance.gaer))
        print("        + aer                = {}".format(account.balance.aer))
        print("    - nonce                  = {}".format(account.nonce))
        print("    - code hash              = {}".format(account.code_hash))
        print("    - storage root           = {}".format(account.storage_root))
        print("    - sql recovery point     = {}"
              .format(account.sql_recovery_point))

    print("------ Get Accounts in Node -----------")
    accounts = aergo.get_node_accounts()
    for i, account in enumerate(accounts):
        print("  > account state : {}".format(account))
        print("    - balance")
        print("        + aergo              = {}"
              .format(account.balance.aergo))
        print("        + gaer               = {}".format(account.balance.gaer))
        print("        + aer                = {}".format(account.balance.aer))
        print("    - nonce                  = {}".format(account.nonce))
        print("    - code hash              = {}".format(account.code_hash))
        print("    - storage root           = {}".format(account.storage_root))
        print("    - sql recovery point     = {}"
              .format(account.sql_recovery_point))

    print("------ Create Account -----------")
    account = aergo.new_account()

    print("Private Key      = {}".format(account.private_key))
    print("Public Key       = {}".format(account.public_key))
    print("Address          = {}".format(account.address))

    print("------ Get Account State -----------")
    aergo.get_account()
    print("  > account state in 'aergo'")
    print("    - balance")
    print("        + aergo              = {}".format(account.balance.aergo))
    print("        + gaer               = {}".format(account.balance.gaer))
    print("        + aer                = {}".format(account.balance.aer))
    print("    - nonce                  = {}".format(account.nonce))
    print("    - code hash              = {}".format(account.code_hash))
    print("    - storage root           = {}".format(account.storage_root))
    print("    - sql recovery point     = {}"
          .format(account.sql_recovery_point))
    print(account)

    print("------ Get Configured Account -----------")
    accounts = [
        {
            "private_key":
                "eHoEcHnaxpGpgzknXjuwon8VFVrLkKHC4FckGuGkQ8depiDDfyUAWC3L",
            "address": "AmPZKCJpT98V9Tc8dBUbRg78M1jgoB1ZEh97Rs1r5KewPcCiURf7",
        },
    ]
    for i, account in enumerate(accounts):
        print("  [{}]".format(i))
        print("    > private key   : {}".format(account['private_key']))
        print("    > address       : {}".format(account['address']))

        # check account state
        a = aergo.get_account(address=account['address'])
        print("    > account state : {}".format(a))
        print("    - balance")
        print("        + aergo              = {}".format(a.balance.aergo))
        print("        + gaer               = {}".format(a.balance.gaer))
        print("        + aer                = {}".format(a.balance.aer))
        print("    - nonce                  = {}".format(a.nonce))
        print("    - code hash              = {}".format(a.code_hash))
        print("    - storage root           = {}".format(a.storage_root))
        print("    - sql recovery point     = {}".format(a.sql_recovery_point))



import sys
import traceback
import time
import json

import aergo.herapy as herapy


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    try:
        print("------ Set Sender Account -----------")
        sender_json = '{"address": "AmNW9YMMm48jTxX5Yee6tRYLpuptWdg3cqbD2CVoee1YcUGBHfad", "enc_key": "47QHSuhBTZ4b7nbw2zJz5EARyQ4XS8gDxhFDeu5mYeKNqfHhwRutUmja3WRxV1suB12eWBeDZ"}'
        sender_account = herapy.Account.from_json(sender_json, password='5678')
        print("  > Sender Account before connecting: {}".format(sender_account))
        print("     > address:     {}".format(sender_account.address))
        print("     > private key: {}".format(sender_account.private_key))

        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Set Sender Account -----------")
        aergo.account = sender_account
        aergo.get_account()
        print("  > Sender Account after connecting: {}".format(aergo.account))

        print("------ Set Receiver Account -----------")
        receiver_json = '{"address": "AmMPQqRJ4pd8bS9xdkw5pjExtLjaUXaYGB5kagzHu4A9ckKgnBV2"}'
        receiver_account = herapy.Account.from_json(receiver_json)
        receiver_account = aergo.get_account(account=receiver_account)
        print("  > Receiver Account: {}".format(receiver_account))

        print("------ Simple Transfer -----------")
        simple_tx, result = aergo.transfer(to_address=receiver_account.address,
                                           amount=10000000000)
        print("  > simple TX[{}]".format(simple_tx.calculate_hash()))
        print("{}".format(str(simple_tx)))
        print("  > result: ", result)
        if result.status != herapy.CommitStatus.TX_OK:
            eprint("    > ERROR[{0}]: {1}".format(result.status, result.detail))
            aergo.disconnect()
            return
        else:
            print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))

        print("------ Check Account Info #1 -----------")
        aergo.get_account()
        print("  > Sender Account: {}".format(aergo.account))
        receiver_account = aergo.get_account(receiver_account)
        print("  > Receiver Account: {}".format(receiver_account))

        time.sleep(3)

        print("------ Check Account Info #2 -----------")
        aergo.get_account()
        print("  > Sender Account: {}".format(aergo.account))
        receiver_account = aergo.get_account(receiver_account)
        print("  > Receiver Account: {}".format(receiver_account))

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



import aergo.herapy as herapy


def test_account_json(aergo) -> None:
    print("------ Set Sender Account -----------")
    sender_json = '{' \
        '"address": "AmNW9YMMm48jTxX5Yee6tRYLpuptWdg3cqbD2CVoee1YcUGBHfad",'\
        '"enc_key": "47QHSuhBTZ4b7nbw2zJz5EARyQ4XS8gDxhFDeu5mYeKNqfHhwRutU' \
        'mja3WRxV1suB12eWBeDZ"}'
    sender_account = herapy.Account.from_json(sender_json, password='5678')
    print("  > Sender Account before connecting: {}".format(sender_account))
    print("     > address:     {}".format(sender_account.address))
    print("     > private key: {}".format(sender_account.private_key))

    print("------ Set Sender Account -----------")
    aergo.account = sender_account
    aergo.get_account()
    print("  > Sender Account after connecting: {}".format(aergo.account))

    print("------ Set Receiver Account -----------")
    receiver_json = \
        '{"address": "AmMPQqRJ4pd8bS9xdkw5pjExtLjaUXaYGB5kagzHu4A9ckKgnBV2"}'
    receiver_account = herapy.Account.from_json(receiver_json)
    receiver_account = aergo.get_account(account=receiver_account)
    print("  > Receiver Account: {}".format(receiver_account))

    print("------ Simple Transfer -----------")

    print("------ Check Account Info Before -----------")
    aergo.get_account()
    print("  > Sender Account: {}".format(aergo.account))
    receiver_account = aergo.get_account(receiver_account)
    print("  > Receiver Account: {}".format(receiver_account))
    receiver_balance_before = int(receiver_account.balance)

    amount = 10000000000
    simple_tx, result = aergo.transfer(to_address=receiver_account.address,
                                       amount=amount)
    print("  > simple TX[{}]".format(simple_tx.calculate_hash()))
    print("{}".format(str(simple_tx)))
    print("  > result: ", result)
    assert result.status == herapy.CommitStatus.TX_OK, \
        "Tx didn't commit: {}, {}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))

    aergo.wait_tx_result(simple_tx.tx_hash)

    print("------ Check Account Info After -----------")
    aergo.get_account()
    print("  > Sender Account: {}".format(aergo.account))
    receiver_account = aergo.get_account(receiver_account)
    print("  > Receiver Account: {}".format(receiver_account))

    assert int(receiver_account.balance) == \
        int(receiver_balance_before) + amount



import aergo.herapy as herapy
import datetime
import asyncio


def test_sc_batch(aergo) -> None:
    print("------ Payload -----------")
    """
-- 'payload' is compiled by aergoluac from
function setItem(key, value)
    system.print('setItem: key='..key..', value='..value)
    system.setItem(key, value)
end

function getItem(key)
    system.print('getItem: key='..key)
    return system.getItem(key)
end

abi.register(setItem, getItem)
    """

    payload_str = "246oyQ629vXweGRbeS8r29poBjGsiqJL23BBaYh8TaSgvGhMAXHrk9vUX" \
        "etphe6m9xCPrRDpQH3fHap4SZUpNWartsHTnJxQXp2zs5mMwK5cwiCgSCePwefA1U4f" \
        "wyt94Q249MxQ2evTKc7vcUVjjn6AnKCqMJTGCgVtnLrpeVMiVEcYBszJtprA3YQmgA8" \
        "vsnfcw5ocw4149wDhV8drF4zviYJSfhGP5TRj7NhLLBstiktq2DzPW8yCrVu2fEnyzf" \
        "Jc1yY9ENuu2K7LyxC7arqGTafZEzsQrrfNo63PXfsSq6oSW7Ev89Y3cFfRjpQfr8mHZ" \
        "PmfCRykUF7aepvkH64r6S7Sg3uUpgrZCnDKuDRuBCizBHUxJDUCua3tJNA2dKFtx9dp" \
        "STfPFkdvPoBFsTjegu6DwHgDop5zZMPV5S4Q5oi5vB1UsvSoX81cMAfp38kJe45ko9r" \
        "LZKk83zpSfgzMLnjDZWccmjtNdnHFhUp9KZG3JRZnr7QRmAHwvTYLd83AS9uCVJhwg5" \
        "oGhVNgMeTnN2f1xZ9Gd9C5NfJTvwYHPVysD7HRwyXVTwYgQhjxwTrxkGcz54gJzF2eT" \
        "DqWnEQKZN61322SZG"

    payload = herapy.utils.decode_address(payload_str)
    print(''.join('{:d} '.format(x) for x in payload))

    byte_code_len = int.from_bytes(payload[:4], byteorder='little')
    print("payload: byte code length = ", byte_code_len)

    print("------ Set Sender Account -----------")
    sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
    sender_account = aergo.new_account(private_key=sender_private_key)
    print("  > Sender Address: {}".format(sender_account.address))
    print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))

    print("------ Deploy SC -----------")
    tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234)
    print("  > TX: {}".format(tx.tx_hash))
    print("{}".format(herapy.utils.convert_tx_to_json(tx)))
    assert result.status == herapy.CommitStatus.TX_OK, \
        "    > ERROR[{0}]: {1}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
    print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

    print("------ Check deployment of SC -----------")
    result = {}
    asyncio.run(aergo.aio_wait_tx_result(tx.tx_hash, result=result))
    aio_result = result.get('tx_result')
    print("COROUTINE tx result: {}".format(str(aio_result)))

    result = aergo.wait_tx_result(tx.tx_hash)
    print("MAIN THREAD tx result: {}".format(str(result)))

    assert aio_result.tx_id == result.tx_id
    assert aio_result.contract_address == result.contract_address
    assert aio_result.status == result.status

    sc_address = result.contract_address
    print("  > SC Address: {}".format(sc_address))

    print("------ Batch Call SC -----------")
    time_start = datetime.datetime.now()
    nonce = aergo.account.nonce + 1
    total_tx_cnt = 1000
    sc_txs = []
    for i in range(total_tx_cnt):
        sc_txs.append(aergo.new_call_sc_tx(sc_address, "setItem",
                                           args=["key{}".format(i),
                                                 "value{}".format(i)],
                                           nonce=nonce + i))
    aergo.batch_call_sc(sc_txs)

    print("let's wait for all results at once")
    results = aergo.await_batch_result(sc_txs)

    assert len(sc_txs) == len(results)

    for i, result in enumerate(results):
        tx = sc_txs[i]
        print("  > TX[{0}] Result: {1}".format(i, str(tx.tx_hash)))
        # print("  > TX[{0}] {1}".format(i, str(tx.tx_hash)))
        # print("    Result: {}".format(str(result)))
        assert result.status == herapy.TxResultStatus.SUCCESS, \
            "  > ERROR[{0}]:{1}: {2}".format(
                result.contract_address, result.status, result.detail)
        assert sc_address == result.contract_address

    time_diff = datetime.datetime.now() - time_start
    time_diff_min = int(time_diff.seconds / 60)
    time_diff_sec = int(time_diff.seconds % 60)
    time_diff_mil = int(time_diff.microseconds / 1000)
    time_diff_mic = int(time_diff.microseconds % 1000)
    print("Execution time: {} = "
          "{} min, {} secs, "
          "{} milli, {} micro".format(time_diff.total_seconds(),
                                      time_diff_min,
                                      time_diff_sec,
                                      time_diff_mil,
                                      time_diff_mic))

    print("------ Query SC -----------")
    for i in range(total_tx_cnt):
        key = "key{}".format(i)
        value = "\"value{}\"".format(i).encode()
        assert value == aergo.query_sc(sc_address, "getItem", args=[key])



import pytest
import aergo.herapy as herapy


def test_blockchain_cert() -> None:
    pytest.skip("Test me with a locally generated certificate and keys")
    aergo = herapy.Aergo()

    print("------ Connect AERGO -----------")
    aergo.connect('localhost:7846',
                  tls_cert="./tests/test_integration/cert/client.pem",
                  tls_ca_cert="./tests/test_integration/rootca/rootca.crt",
                  tls_key="./tests/test_integration/cert/client.key")

    print("------ Get Node Info -----------")
    node_state = aergo.get_node_state()
    print("Node State: {}".format(str(node_state)))
    node_info = aergo.get_node_info()
    print("Node Info: {}".format(str(node_info)))

    print("------ Get Blockchain Info -----------")
    blockchain_info = aergo.get_chain_info()
    print("Blockchain Info: {}".format(str(blockchain_info)))
    blockchain_info = aergo.get_chain_info(with_consensus_info=False)
    print("Blockchain Info: {}".format(str(blockchain_info)))

    print("------ Get Consensus Info -----------")
    consensus_info = aergo.get_consensus_info()
    print("Consensus Info: {}".format(str(consensus_info)))



# -*- coding: utf-8 -*-

import ecdsa
import hashlib
import json

from google.protobuf.json_format import (
    MessageToJson,
    Parse
)
from cryptography.exceptions import InvalidTag
from typing import (
    Union,
    Dict,
    Optional
)

from .grpc import blockchain_pb2

from .obj import address as addr
from .obj import aer
from .obj import private_key as pk

from .utils.encoding import (
    decode_root,
    encode_private_key,
    decode_private_key
)
from .utils import merkle_proof as mp
from .utils.encryption import (
    encrypt_bytes,
    decrypt_bytes,
    decrypt_keystore_v1,
    encrypt_to_keystore_v1
)

from .errors.general_exception import GeneralException


class Account:
    """ Account can be a user account with private and public key,
    or a contract account.
    """
    def __init__(
        self,
        private_key: Union[str, bytes, None] = None,
        empty: bool = False
    ) -> None:
        if empty:
            self.__private_key = None
            self.__address = None
            self.__state = None
            self.__state_proof = None
            return

        self.__private_key = pk.PrivateKey(private_key)
        self.__address = self.__private_key.address
        self.__state = None
        self.__state_proof = None

    def json(
        self,
        password: Union[bytes, str, None] = None,
        with_private_key: bool = False
    ) -> Dict:
        state = self.state
        is_state_proof = False
        if self.__state_proof:
            state = MessageToJson(self.__state_proof)
            is_state_proof = True

        if state:
            state = json.loads(state)

        account = {
            "address": str(self.__address),
            "balance": str(self.balance),
            "nonce": str(self.nonce),
            "state": state,
            "is_state_proof": is_state_proof,
        }

        if password and self.__private_key is not None:
            enc_key = Account.encrypt_account(self, password)
            account["enc_key"] = encode_private_key(enc_key)

        if with_private_key and self.__private_key is not None:
            account["priv_key"] = str(self.__private_key)

        return account

    @staticmethod
    def from_json(
        data: Union[Dict, str],
        password: Union[bytes, str, None] = None
    ) -> 'Account':
        if isinstance(data, str):
            data_dict: Dict = json.loads(data)
            data = data_dict

        # handle private key or encrypted key
        priv_key = data.get('priv_key', None)
        enc_key = data.get('enc_key', None)
        if priv_key:
            account = Account(private_key=priv_key)
        elif enc_key:
            enc_key = decode_private_key(enc_key)
            if not password:
                raise GeneralException("Must provide a password to decrypt")
            account = Account.decrypt_account(enc_key, password)
        else:
            account = Account(empty=True)

        # handle address
        address = data.get('address', None)
        if address is None:
            raise ValueError("address cannot be None")
        elif account.address is not None and address != str(account.address):
            raise ValueError("address is irrelevant with the private key")
        elif account.address is None:
            account.address = address

        # handle state of state_proof
        state = data.get('state', None)
        if state:
            state = json.dumps(state)
            is_state_proof = data.get('is_state_proof')
            if is_state_proof:
                state_proof_pb2 = blockchain_pb2.AccountProof()
                Parse(state, state_proof_pb2, ignore_unknown_fields=True)
                account.state_proof = state_proof_pb2
            else:
                state_pb2 = blockchain_pb2.State()
                Parse(state, state_pb2, ignore_unknown_fields=True)
                account.state = state_pb2

        return account

    @staticmethod
    def decrypt_from_keystore(
        keystore: Union[Dict, str],
        password: str
    ) -> 'Account':
        if isinstance(keystore, str):
            keystore_dict: Dict = json.loads(keystore)
            keystore = keystore_dict
        try:
            privkey_raw = decrypt_keystore_v1(keystore, password)
        except AssertionError as e:
            raise GeneralException("Failed to decrypt keystore") from e
        return Account(private_key=privkey_raw)

    @staticmethod
    def encrypt_to_keystore(
        account: 'Account',
        password: str,
        kdf_n: int = 2**18
    ) -> Dict:
        if not account.private_key:
            raise GeneralException("Account must have a private key")
        try:
            keystore = encrypt_to_keystore_v1(
                bytes(account.private_key),
                str(account.address),
                password, kdf_n=kdf_n
            )
        except AssertionError as e:
            raise GeneralException("Failed to encrypt keystore") from e
        return keystore

    def __str__(self) -> str:
        return json.dumps(self.json(), indent=2)

    def sign_msg_hash(self, msg_hash: bytes) -> Optional[bytes]:
        return self.__private_key.sign_msg(msg_hash) if \
            self.__private_key else None

    def verify_sign(self, msg_hash: bytes, sign: bytes) -> Optional[bool]:
        return self.__private_key.verify_sign(msg_hash, sign) if \
            self.__private_key else None

    @property
    def private_key(self) -> Optional[pk.PrivateKey]:
        return self.__private_key

    @property
    def public_key(self) -> Optional[ecdsa.ecdsa.Public_key]:
        if self.__private_key is not None:
            return self.__private_key.public_key

        if self.__address is not None:
            return self.__address.public_key

        return None

    @property
    def address(self) -> Optional[addr.Address]:
        return self.__address

    @address.setter
    def address(self, v: Union[str, bytes]) -> None:
        if self.__address is not None:
            raise ValueError('not empty account')

        self.__address = addr.Address(None, empty=True)
        self.__address.value = v  # type: ignore

    @property
    def state(self) -> Optional[str]:
        if self.__state is None:
            return None
        return MessageToJson(self.__state)

    @state.setter
    def state(self, v):
        self.__state = v

    @property
    def state_proof(self):
        return self.__state_proof

    @state_proof.setter
    def state_proof(self, v):
        self.__state_proof = v

    @property
    def nonce(self):
        if self.__state is None:
            return -1
        return self.__state.nonce

    @nonce.setter
    def nonce(self, v):
        if self.__state.nonce > v:
            raise ValueError(
                "the nonce value should be bigger than the current nonce "
                "value: {}".format(self.__state.nonce)
            )
        self.__state.nonce = v

    @property
    def balance(self):
        if self.__state is None:
            return aer.Aer()
        return aer.Aer(int.from_bytes(self.__state.balance, 'big'))

    @property
    def code_hash(self):
        if self.__state is None:
            return None
        return self.__state.codeHash

    @property
    def storage_root(self):
        if self.__state is None:
            return None
        return self.__state.storageRoot

    @property
    def sql_recovery_point(self):
        if self.__state is None:
            return None
        return self.__state.sqlRecoveryPoint

    @staticmethod
    def encrypt_account(
        account: 'Account',
        password: Union[str, bytes]
    ) -> bytes:
        """
        https://cryptography.io/en/latest/hazmat/primitives/aead/
        :param account: account to export
        :return: encrypted account data (bytes)
        """
        if not account.private_key:
            raise GeneralException("Account must have a private key")
        return encrypt_bytes(bytes(account.private_key), password)

    @staticmethod
    def decrypt_account(
        encrypted_bytes: bytes,
        password: Union[str, bytes]
    ) -> 'Account':
        """
        https://cryptography.io/en/latest/hazmat/primitives/aead/
        :param encrypted_bytes: encrypted data (bytes) of account
        :param password: to decrypt the exported bytes
        :return: account instance
        """
        try:
            dec_value = decrypt_bytes(encrypted_bytes, password)
        except InvalidTag as e:
            raise GeneralException(
                "Fail to decrypt an account. Please check the password."
            ) from e
        return Account(private_key=dec_value)

    def verify_proof(self, root: Union[str, bytes]) -> bool:
        """ verify that the given inclusion and exclusion proofs are
            correct
        """
        if self.__state_proof is None:
            return False

        if isinstance(root, str) and len(root) != 0:
            root = decode_root(root)

        if bytes(self.__address) != self.__state_proof.key:
            return False

        trie_key = hashlib.sha256(bytes(self.address)).digest()
        value = hashlib.sha256(
            self.__state_proof.state.SerializeToString()).digest()
        ap = self.__state_proof.auditPath

        if self.__state_proof.bitmap:
            height = self.__state_proof.height
            bitmap = self.__state_proof.bitmap
            if self.__state_proof.inclusion:
                return mp.verify_inclusion_c(
                    ap, height, bitmap, root, trie_key, value)
            else:
                return mp.verify_exclusion_c(
                    root, ap, height, bitmap, trie_key,
                    self.__state_proof.proofKey, self.__state_proof.proofVal
                )
        else:
            if self.__state_proof.inclusion:
                return mp.verify_inclusion(ap, root, trie_key, value)
            else:
                return mp.verify_exclusion(
                    root, ap, trie_key, self.__state_proof.proofKey,
                    self.__state_proof.proofVal
                )



ADDRESS_BYTES_LENGTH = 33
ADDRESS_VERSION = b'\x42'
CONTRACT_VERSION = b'\xC0'

PRIVATE_KEY_VERSION = b'\xAA'

PUBLIC_KEY_UNCOMPRESSED = b'\x04'
PUBLIC_KEY_COMPRESSED_E = b'\x02'
PUBLIC_KEY_COMPRESSED_O = b'\x03'



def test_blockchain_stream(aergo) -> None:
    print("------ Get Block Meta Stream -----------")
    stream = aergo.receive_block_stream()
    i = 0
    while i < 3:
        block = next(stream)
        print('[{}] block: {}'.format(i, str(block)))
        i += 1

    print("------ Get Block Stream -----------")
    stream = aergo.receive_block_stream()
    i = 0
    while i < 3:
        block = next(stream)
        print('[{}] block: {}'.format(i, str(block)))
        i += 1



# -*- coding: utf-8 -*-

"""Communication(grpc) module."""

import grpc
from typing import (
    Optional,
    List
)

from .grpc import account_pb2, blockchain_pb2, rpc_pb2, rpc_pb2_grpc, raft_pb2
from .utils.converter import convert_tx_to_grpc_tx
from .obj.transaction import Transaction


class Comm:
    def __init__(
        self,
        target: Optional[str] = None,
        tls_ca_cert: Optional[bytes] = None,
        tls_cert: Optional[bytes] = None,
        tls_key: Optional[bytes] = None
    ):
        self.__target = target
        self.__tls_ca_cert = tls_ca_cert
        self.__tls_cert = tls_cert
        self.__tls_key = tls_key
        self.__channel = None
        self.__rpc_stub = None

    def connect(self):
        self.disconnect()

        if self.__tls_cert is not None:
            creds = grpc.ssl_channel_credentials(
                root_certificates=self.__tls_ca_cert,
                private_key=self.__tls_key, certificate_chain=self.__tls_cert
            )
            self.__channel = grpc.secure_channel(self.__target, creds)
        else:
            self.__channel = grpc.insecure_channel(self.__target)

        self.__rpc_stub = rpc_pb2_grpc.AergoRPCServiceStub(self.__channel)

    def disconnect(self):
        if self.__channel is not None:
            self.__channel.close()

    def create_account(self, address: bytes, passphrase: str):
        if self.__rpc_stub is None:
            return None
        account = account_pb2.Account(address=address)
        return self.__rpc_stub.CreateAccount(
            request=rpc_pb2.Personal(account=account, passphrase=passphrase)
        )

    def get_account_state(self, address: bytes):
        if self.__rpc_stub is None:
            return None

        rpc_account = account_pb2.Account()
        rpc_account.address = address
        return self.__rpc_stub.GetState(rpc_account)

    def get_account_state_proof(
        self,
        address: bytes,
        root: bytes,
        compressed: bool
    ):
        if self.__rpc_stub is None:
            return None
        account_and_root = rpc_pb2.AccountAndRoot(Account=address,
                                                  Root=root,
                                                  Compressed=compressed)
        return self.__rpc_stub.GetStateAndProof(account_and_root)

    def get_chain_info(self):
        if self.__rpc_stub is None:
            return None

        return self.__rpc_stub.GetChainInfo(rpc_pb2.Empty())

    def get_consensus_info(self):
        if self.__rpc_stub is None:
            return None

        return self.__rpc_stub.GetConsensusInfo(rpc_pb2.Empty())

    def get_node_info(self, keys: Optional[str]):
        if self.__rpc_stub is None:
            return None

        # currently not working
        key_params = rpc_pb2.KeyParams()
        if keys is not None:
            key_params.key.extend(keys)

        return self.__rpc_stub.GetServerInfo(key_params)

    def receive_event_stream(
        self,
        sc_address: bytes,
        event_name: str,
        start_block_no: int,
        end_block_no: int,
        with_desc: bool,
        arg_filter: Optional[bytes],
        recent_block_cnt: int
    ):
        if self.__rpc_stub is None:
            return None

        filter = blockchain_pb2.FilterInfo()
        filter.contractAddress = sc_address
        filter.eventName = event_name
        filter.blockfrom = start_block_no
        filter.blockto = end_block_no
        filter.desc = with_desc
        if arg_filter is not None:
            filter.argFilter = arg_filter
        filter.recentBlockCnt = recent_block_cnt

        return self.__rpc_stub.ListEventStream(filter)

    def get_events(
        self,
        sc_address: bytes,
        event_name: str,
        start_block_no: int,
        end_block_no: int,
        with_desc: bool,
        arg_filter: Optional[bytes],
        recent_block_cnt: int
    ):
        if self.__rpc_stub is None:
            return None

        filter = blockchain_pb2.FilterInfo()
        filter.contractAddress = sc_address
        filter.eventName = event_name
        if start_block_no > 0:
            filter.blockfrom = start_block_no
        if end_block_no > 0:
            filter.blockto = end_block_no
        filter.desc = with_desc
        if arg_filter is not None:
            filter.argFilter = arg_filter
        filter.recentBlockCnt = recent_block_cnt

        return self.__rpc_stub.ListEvents(filter)

    def receive_block_meta_stream(self):
        if self.__rpc_stub is None:
            return None
        return self.__rpc_stub.ListBlockMetadataStream(rpc_pb2.Empty())

    def receive_block_stream(self):
        if self.__rpc_stub is None:
            return None
        return self.__rpc_stub.ListBlockStream(rpc_pb2.Empty())

    def get_block_headers(
        self,
        block_hash: Optional[bytes],
        block_height: int,
        list_size: int,
        offset: int,
        is_asc_order: bool
    ):
        if self.__rpc_stub is None:
            return None
        params = rpc_pb2.ListParams()
        if block_hash is not None:
            params.hash = block_hash
        if block_height >= 0:
            params.height = block_height
        params.size = list_size
        params.offset = offset
        params.asc = is_asc_order
        return self.__rpc_stub.ListBlockHeaders(params)

    def get_block_metas(
        self,
        block_hash: Optional[bytes],
        block_height: int,
        list_size: int,
        offset: int,
        is_asc_order: bool
    ):
        if self.__rpc_stub is None:
            return None
        params = rpc_pb2.ListParams()
        if block_hash is not None:
            params.hash = block_hash
        if block_height >= 0:
            params.height = block_height
        params.size = list_size
        params.offset = offset
        params.asc = is_asc_order
        return self.__rpc_stub.ListBlockMetadata(params)

    def get_blockchain_status(self):
        if self.__rpc_stub is None:
            return None

        return self.__rpc_stub.Blockchain(rpc_pb2.Empty())

    def get_accounts(self):
        if self.__rpc_stub is None:
            return None
        return self.__rpc_stub.GetAccounts(rpc_pb2.Empty())

    def get_block(self, query: bytes):
        if self.__rpc_stub is None:
            return None
        v = rpc_pb2.SingleBytes()
        v.value = query
        return self.__rpc_stub.GetBlock(v)

    def get_block_meta(self, query: bytes):
        if self.__rpc_stub is None:
            return None
        v = rpc_pb2.SingleBytes()
        v.value = query
        return self.__rpc_stub.GetBlockMetadata(v)

    def get_peers(self):
        if self.__rpc_stub is None:
            return None
        return self.__rpc_stub.GetPeers(rpc_pb2.Empty())

    def get_node_state(self, timeout: int):
        if self.__rpc_stub is None:
            return None
        single_bytes = rpc_pb2.SingleBytes()
        single_bytes.value = timeout.to_bytes(8, byteorder='little')
        return self.__rpc_stub.NodeState(single_bytes)

    def get_tx(self, tx_hash: bytes):
        if self.__rpc_stub is None:
            return None
        single_bytes = rpc_pb2.SingleBytes()
        single_bytes.value = tx_hash
        return self.__rpc_stub.GetTX(single_bytes)

    def get_block_tx(self, tx_hash: bytes):
        if self.__rpc_stub is None:
            return None
        single_bytes = rpc_pb2.SingleBytes()
        single_bytes.value = tx_hash
        return self.__rpc_stub.GetBlockTX(single_bytes)

    def unlock_account(self, address: bytes, passphrase: str):
        if self.__rpc_stub is None:
            return None
        account = account_pb2.Account(address=address)
        personal = rpc_pb2.Personal(passphrase=passphrase, account=account)
        return self.__rpc_stub.UnlockAccount(request=personal)

    def lock_account(self, address: bytes, passphrase: str):
        if self.__rpc_stub is None:
            return None
        account = account_pb2.Account(address=address)
        personal = rpc_pb2.Personal(passphrase=passphrase, account=account)
        return self.__rpc_stub.LockAccount(request=personal)

    # This RPC is for making and sending Tx inside a node.
    # Don't use it for sending TX which is made by a client.
    def send_tx(self, unsigned_tx: Transaction):
        if self.__rpc_stub is None:
            return None
        return self.__rpc_stub.SendTX(convert_tx_to_grpc_tx(unsigned_tx))

    # This RPC is for sending signed Txs made by a client.
    def commit_txs(self, signed_txs: List[Transaction]):
        if self.__rpc_stub is None:
            return None
        rpc_txs = []
        for signed_tx in signed_txs:
            rpc_txs.append(convert_tx_to_grpc_tx(signed_tx))
        rpc_tx_list = blockchain_pb2.TxList()
        rpc_tx_list.txs.extend(rpc_txs)
        return self.__rpc_stub.CommitTX(rpc_tx_list)

    def get_receipt(self, tx_hash: bytes):
        if self.__rpc_stub is None:
            return None
        v = rpc_pb2.SingleBytes()
        v.value = tx_hash
        return self.__rpc_stub.GetReceipt(v)

    def query_contract(self, sc_address: bytes, query_info: bytes):
        query = blockchain_pb2.Query()
        if self.__rpc_stub is None:
            return None
        query.contractAddress = sc_address
        query.queryinfo = query_info
        return self.__rpc_stub.QueryContract(query)

    def query_contract_state(
        self,
        sc_address: bytes,
        storage_keys: List[bytes],
        root: bytes,
        compressed: bool
    ):
        if self.__rpc_stub is None:
            return None
        state_query = blockchain_pb2.StateQuery(contractAddress=sc_address,
                                                storageKeys=storage_keys,
                                                root=root,
                                                compressed=compressed)
        return self.__rpc_stub.QueryContractState(state_query)

    def get_conf_change_progress(self, block_height: int):
        if self.__rpc_stub is None:
            return None

        v = rpc_pb2.SingleBytes()
        v.value = block_height.to_bytes(8, byteorder='little')
        return self.__rpc_stub.GetConfChangeProgress(v)

    def get_enterprise_config(self, key: str):
        if self.__rpc_stub is None:
            return None

        v = rpc_pb2.EnterpriseConfigKey()
        v.key = key
        return self.__rpc_stub.GetEnterpriseConfig(v)

    def get_name_info(self, name: str, block_height: int):
        if self.__rpc_stub is None:
            return None

        n = rpc_pb2.Name()
        n.name = name
        n.blockNo = block_height
        return self.__rpc_stub.GetNameInfo(n)

    def get_abi(self, addr_bytes: bytes):
        if self.__rpc_stub is None:
            return None
        v = rpc_pb2.SingleBytes()
        v.value = addr_bytes
        return self.__rpc_stub.GetABI(v)

    def add_raft_member(
        self,
        request_id: int,
        member_id: int,
        member_name: str,
        member_address: str,
        member_peer_id: bytes
    ):
        if self.__rpc_stub is None:
            return None
        ch = raft_pb2.MembershipChange()
        ch.type = raft_pb2.ADD_MEMBER
        ch.requestID = request_id
        ch.attr = raft_pb2.MemberAttr()
        ch.attr.ID = member_id
        ch.attr.name = member_name
        ch.attr.address = member_address
        ch.attr.peerID = member_peer_id
        return self.__rpc_stub.ChangeMembership(ch)

    def del_raft_member(
        self,
        request_id: int,
        member_id: int,
        member_name: str,
        member_address: str,
        member_peer_id: bytes
    ):
        if self.__rpc_stub is None:
            return None
        ch = raft_pb2.MembershipChange()
        ch.type = raft_pb2.REMOVE_MEMBER
        ch.requestID = request_id
        ch.attr = raft_pb2.MemberAttr()
        ch.attr.ID = member_id
        ch.attr.name = member_name
        ch.attr.address = member_address
        ch.attr.peerID = member_peer_id
        return self.__rpc_stub.ChangeMembership(ch)



import time

from threading import Thread

import aergo.herapy as herapy


def read_movie_event_stream(movie_event_stream) -> None:
    while True:
        try:
            print("RECEIVE: 'movie' Event: {}"
                  .format(str(next(movie_event_stream))))
            print(movie_event_stream.is_active())
            print(movie_event_stream.cancelled())
            print(movie_event_stream.done())
            print(movie_event_stream.running())
        except Exception as e:
            print(e)
            print(movie_event_stream.is_active())
            print(movie_event_stream.cancelled())
            print(movie_event_stream.done())
            print(movie_event_stream.running())
            break

    print("END 'movie' event stream")
    movie_event_stream.stop()


def test_sc_event_stream(aergo) -> None:
    print("------ Payload -----------")
    """

function constructor(name)
        contract.event("movie", "robocop", 1, "I'll be back")
end

function movie_event(name)
        contract.event("movie", "arnold", 2, "It's mine")
        contract.event("movie", "terminator", 3, "No, it's mine")
end

function ani_event(name)
        contract.event("a-time!!")
        contract.event("a-time!!", "pinn", "Adventure time!!", false,
                       bignum.number("999999999999999"),
                       {"princess bubble gum", 1}
        )
        contract.event("a-time!!", "Jake", "Adventure time!!", false,
                       {"whatever", 1, {true, false},
                       bignum.number("999999999999999")}
        )

        contract.event("movie", "yoda", 3, "i'm not the ani character.")
end

abi.register(movie_event, ani_event)
    """
    payload_str = "D1ZsfUyj4q8GA6AtbW81TzSTSvqaeUvcsePsaZufptRK5qSb83fyXRWdE" \
        "gGhLWs17JfRNrM4iLjGe54fkyvsyjHVgFCHqTzJWK4Yz8uQwRqzF5x1g1twZ3jPP1Xt" \
        "J95y43RD48qXGoeeoDsmmNErQJCux6d4eLZwE344bz5hM1DdWEd22yE6ogC37RLRtRZ" \
        "5FFKm3HkxaM67G7VqWfsX5t2Tmd23Myn62hkAdNWjSBaWizqyi6ENB3VD2YJGG82bVe" \
        "q9a4zMQdK58Wxnfcfqo14Vo9Y9jsY8A5QYNRi6uotRfiYt3LFrPjPBWgyQRn4qFQypS" \
        "W5vfgFZvv9VBdFPGgoMN5zeBp7VNWXvE5o8CGAf3fuFu2gCGLS4DfCtQe7VNnsCUi7U" \
        "1AdmdupqoxMHM6kCfEtk77YyiaaSTm4xUSt7R2hwzs2Xnfh8CLxMaH2cg3rsWF4fefs" \
        "5KvnjN3F1q43og6BueK66o8NssyY41kAtincbYhDC8oHaBkMTdXD6mkhb3QECwn8dh8" \
        "k9a3KTys1tdUzf99nqpK5YrAcZFDjbrmmmtcBYwZg6zdRAkQnJptvr25HzeHMfbfERH" \
        "8P1XzYmjqYynFvUNv4utuWZ7avS2xvJoZph3nSa8Xuhbdw6TFLVCDZVfZB6j8kQbJnb" \
        "1DoEEtvhPHLMP4StcgiE9grGQuduZ5iHYtS8Nk696Fq6Y9gzp19bWD3VKarq1XuueM6" \
        "FNrzCQNY22nov4mxbM8vcxp4axdjmZNWLcdttFycazorxCNbn1YBQFzGPGzstJPN72R" \
        "B9dxV17cb1TuREKTF7mVLLeahgHsY4eVnA1m5Z8U7tFKswRDTGeXv6vCt7MstVBztdd" \
        "4sRfHTXKdFCgfoQk2zwzGS8FEVgxAP3Y4Tip4nuQd8hVyTiaApFmNnkD9byxUNCc44T" \
        "r9LAnRycc239HGpZjWBwsJvmb4jsY2fuEELXfSgQf4yZhnbcPC4DYPQLDRXc8E6DdLz" \
        "ixMk7cAfUZqododk9SWzjB9FSKtzVYWiKuyg2uZoRaxDmrtZpe6pXL8F4pzt2P2HEG6" \
        "jyFn5MT9PBWhkj11vqAjWpdRgNAbTKEb5zk9ZeER9YUAT8etk2VFpAGqmGnv9CTEQr2" \
        "rfDQeNgNYzoy44mNZonhi4n4zo9h2HmrSsatWa8oxbJJYihGs7zD65TiqRfnKg43Xx5" \
        "qD"
    payload = herapy.utils.decode_address(payload_str)
    print(''.join('{:d} '.format(x) for x in payload))

    byte_code_len = int.from_bytes(payload[:4], byteorder='little')
    print("payload: byte code length = ", byte_code_len)

    print("------ Set Sender Account -----------")
    sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
    sender_account = aergo.new_account(private_key=sender_private_key)
    print("  > Sender Address: {}".format(sender_account.address))
    print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

    print("------ Deploy SC -----------")
    tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234)
    print("  > TX: {}".format(tx.tx_hash))
    print("{}".format(herapy.utils.convert_tx_to_json(tx)))
    assert result.status == herapy.CommitStatus.TX_OK, \
        "    > ERROR[{0}]: {1}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
    print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

    aergo.wait_tx_result(tx.tx_hash)

    print("------ Check deployment of SC -----------")
    print("  > TX: {}".format(tx.tx_hash))
    result = aergo.get_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.CREATED, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    sc_address = result.contract_address
    print("  > SC Address: {}".format(sc_address))

    print("------ Get reading events threads -----------")
    movie_event_stream = aergo.receive_event_stream(
        sc_address, "movie", start_block_no=0)
    movie_thread = Thread(target=read_movie_event_stream,
                          args=[movie_event_stream])
    movie_thread.start()

    print("------ Call SC -----------")
    tx, result = aergo.call_sc(sc_address, "movie_event")
    print("movie_event tx hash: {}".format(str(tx.tx_hash)))
    aergo.wait_tx_result(tx.tx_hash)

    print("------ Cancel 'movie' event stream -----------")
    movie_event_stream.cancel()

    while True:
        if movie_event_stream.stopped:
            break
        time.sleep(1)



from aergo.herapy.utils.encoding import (
    decode_private_key, encode_b64, encode_address, encode_b58
)
from aergo.herapy.utils.converter import (
    convert_public_key_to_bytes, bytes_to_int_str
)
from aergo.herapy.account import Account
from aergo.herapy.obj.private_key import PrivateKey
from aergo.herapy.obj.address import Address


# exported key from aergocli with password 1234
# private key: 6i2n6TAtsKBaWSujYFzqkBCiP834j1u3nwmffZb8dxCxDcQaXAb
# address: AmLnc5nhXjL3a3GUzv1Hb44LnGPvhZufDZ8s8oE9kazgbW6FgnUa
exported_key_str = \
    "47Nn73DSkBjHSfxvkJtiR3PACzQEZUN8qbjT41fYUfVEYipRZPwY4TnNRz5ahFHgN5naoJfJz"
exported_key = decode_private_key(exported_key_str)


def test_import() -> None:
    assert exported_key
    acc = Account.decrypt_account(exported_key, '1234')
    assert str(acc.address) == \
        "AmLnc5nhXjL3a3GUzv1Hb44LnGPvhZufDZ8s8oE9kazgbW6FgnUa"
    assert str(acc.private_key) == \
        "6i2n6TAtsKBaWSujYFzqkBCiP834j1u3nwmffZb8dxCxDcQaXAb"

    assert acc.private_key
    assert acc.address == acc.private_key.address
    assert acc.public_key == acc.private_key.public_key
    assert acc.public_key.point == acc.private_key.public_key.point
    assert acc.public_key == acc.address.public_key
    assert acc.public_key.point == acc.address.public_key.point
    assert acc.private_key.public_key == acc.address.public_key
    assert acc.private_key.public_key.point == acc.address.public_key.point

    assert bytes(acc.address) == convert_public_key_to_bytes(acc.public_key)
    assert bytes(acc.address) == \
        convert_public_key_to_bytes(acc.private_key.public_key)
    assert bytes(acc.address) == \
        convert_public_key_to_bytes(acc.address.public_key)
    assert bytes(acc.private_key.address) == \
        convert_public_key_to_bytes(acc.public_key)

    private_key = PrivateKey(
        pk="6i2n6TAtsKBaWSujYFzqkBCiP834j1u3nwmffZb8dxCxDcQaXAb")
    assert acc.public_key == private_key.public_key
    assert acc.public_key.point == private_key.public_key.point
    assert private_key.public_key == acc.address.public_key
    assert private_key.public_key.point == acc.address.public_key.point

    assert bytes(acc.address) == \
        convert_public_key_to_bytes(private_key.public_key)
    assert bytes(private_key.address) == \
        convert_public_key_to_bytes(acc.public_key)

    address = Address(None, empty=True)
    address.value = \
        "AmLnc5nhXjL3a3GUzv1Hb44LnGPvhZufDZ8s8oE9kazgbW6FgnUa"  # type: ignore
    assert acc.address != address
    assert bytes(acc.address) == bytes(address)
    assert private_key.address != address
    assert bytes(private_key.address) == bytes(address)

    assert acc.public_key == address.public_key
    assert acc.public_key.point == address.public_key.point
    assert private_key.public_key == address.public_key
    assert private_key.public_key.point == address.public_key.point

    assert bytes(acc.address) == \
        convert_public_key_to_bytes(address.public_key)
    assert bytes(private_key.address) == \
        convert_public_key_to_bytes(address.public_key)


def test_keys() -> None:
    assert exported_key
    acc = Account.decrypt_account(exported_key, '1234')
    assert str(acc.address) == \
        "AmLnc5nhXjL3a3GUzv1Hb44LnGPvhZufDZ8s8oE9kazgbW6FgnUa"

    assert acc.private_key
    signing_key = acc.private_key.get_signing_key()
    verifying_key = signing_key.get_verifying_key()
    print("to pem = {}".format(verifying_key.to_pem()))
    print(
        "int(private key bytes) = {}"
        .format(bytes_to_int_str(bytes(acc.private_key)))
    )

    pubkey_bytes = convert_public_key_to_bytes(acc.public_key)
    print("compressed bytes(public key) = {!r}".format(pubkey_bytes))
    print(
        "compressed int(public key bytes) = {}"
        .format(bytes_to_int_str(pubkey_bytes))
    )

    address = encode_address(pubkey_bytes)
    print("compressed address = {}".format(address))
    assert address == str(acc.address)

    nocomp_pubkey_bytes = convert_public_key_to_bytes(
        acc.public_key, compressed=False)
    print("no compressed bytes(public key) = {!r}".format(pubkey_bytes))
    nocomp_pubkey_txt = encode_b64(nocomp_pubkey_bytes)
    print("no compressed public key = {}".format(nocomp_pubkey_txt))

    address = encode_address(nocomp_pubkey_bytes)
    print("no compressed address = {}".format(address))
    assert address != str(acc.address)

    p2p_pubkey_bytes_head = "\x08\x02\x12".encode("latin-1")
    print(
        "int(p2p public key head bytes) = {}"
        .format(bytes_to_int_str(p2p_pubkey_bytes_head))
    )

    p2p_pubkey_bytes = p2p_pubkey_bytes_head \
        + len(pubkey_bytes).to_bytes(length=1, byteorder='big') \
        + pubkey_bytes
    print("bytes(p2p public key) = {!r}".format(p2p_pubkey_bytes))
    print(
        "int(p2p public key bytes) = {}"
        .format(bytes_to_int_str(p2p_pubkey_bytes))
    )
    pubkey_txt = encode_b64(p2p_pubkey_bytes)
    print("p2p public key = {}".format(pubkey_txt))

    p2p_id_bytes = "\x00\x25".encode("latin-1") + p2p_pubkey_bytes
    print("int(p2p id bytes) = {}".format(bytes_to_int_str(p2p_id_bytes)))
    id = encode_b58(p2p_id_bytes)
    print("p2p id = {}".format(id))


def test_private_key() -> None:
    assert exported_key
    acc = Account.decrypt_account(exported_key, '1234')
    assert str(acc.address) == \
        "AmLnc5nhXjL3a3GUzv1Hb44LnGPvhZufDZ8s8oE9kazgbW6FgnUa"

    assert acc.private_key
    acc2 = Account(private_key=bytes(acc.private_key))
    assert acc.address and acc2.address
    assert bytes(acc.address) == bytes(acc2.address)
    assert str(acc.address) == str(acc2.address)

    acc3 = Account(private_key=str(acc.private_key))
    assert acc.address and acc3.address
    assert bytes(acc.address) == bytes(acc3.address)
    assert str(acc.address) == str(acc3.address)


def test_json() -> None:
    assert exported_key
    acc = Account.decrypt_account(exported_key, '1234')
    assert str(acc.address) == \
        "AmLnc5nhXjL3a3GUzv1Hb44LnGPvhZufDZ8s8oE9kazgbW6FgnUa"

    acc_json = acc.json()
    assert acc_json.get('priv_key', None) is None
    assert acc_json.get('enc_key', None) is None
    assert acc_json.get('address', None) == str(acc.address)
    assert acc_json.get('balance', None) == "0 aer"
    assert acc_json.get('nonce', None) == "-1"
    assert acc_json.get('state', None) is None
    assert acc_json.get('is_state_proof', None) is False

    assert acc.private_key
    acc2 = Account(private_key=bytes(acc.private_key))
    acc2_json = acc2.json(with_private_key=True)
    assert acc2_json.get('priv_key', None) == str(acc.private_key)
    assert acc2_json.get('enc_key', None) is None
    assert acc2_json.get('address', None) == str(acc.address)

    acc2_json2 = acc2.json(password='1234')
    assert acc2_json2.get('priv_key', None) is None
    assert acc2_json2.get('enc_key', None) == exported_key_str
    assert acc2_json2.get('address', None) == str(acc.address)

    acc3 = Account.from_json(acc_json)
    assert acc3.private_key is None
    assert acc.address and acc3.address
    assert bytes(acc.address) == bytes(acc3.address)
    assert str(acc.address) == str(acc3.address)

    acc3_json = acc3.json(with_private_key=True)
    assert acc3_json.get('priv_key', None) is None
    assert acc3_json.get('enc_key', None) is None
    assert acc3_json.get('address', None) == str(acc.address)

    acc3_json2 = acc3.json(password='1234')
    assert acc3_json2.get('priv_key', None) is None
    assert acc3_json2.get('enc_key', None) is None
    assert acc3_json2.get('address', None) == str(acc.address)

    acc4 = Account.from_json(acc2_json)
    assert acc.private_key and acc4.private_key
    assert bytes(acc.private_key) == bytes(acc4.private_key)
    assert str(acc.private_key) == str(acc4.private_key)
    assert acc.address and acc4.address
    assert bytes(acc.address) == bytes(acc4.address)
    assert str(acc.address) == str(acc4.address)
    assert bytes(acc3.address) == bytes(acc4.address)
    assert str(acc3.address) == str(acc4.address)

    acc5 = Account.from_json(acc2_json2, password='1234')
    assert acc.private_key and acc5.private_key
    assert bytes(acc.private_key) == bytes(acc5.private_key)
    assert str(acc.private_key) == str(acc5.private_key)
    assert acc.address and acc5.address
    assert bytes(acc.address) == bytes(acc5.address)
    assert str(acc.address) == str(acc5.address)
    assert bytes(acc3.address) == bytes(acc5.address)
    assert str(acc3.address) == str(acc5.address)


def test_from_json() -> None:
    private_key_str = \
        "47DTHPaRpbJ67KZvBaciz68EJ5k6E2FNBUuknKR6NA4t8oyA3uyvj" \
        "S2QKxJ7JsgHMunPUitoT"
    private_key = decode_private_key(private_key_str)
    assert private_key
    acc = Account.decrypt_account(encrypted_bytes=private_key, password='1234')
    assert str(acc.address) == \
        "AmNW9YMMm48jTxX5Yee6tRYLpuptWdg3cqbD2CVoee1YcUGBHfad"

    acc2_json_str = '{' \
        '"address": "AmNW9YMMm48jTxX5Yee6tRYLpuptWdg3cqbD2CVoee1YcUGBHfad", ' \
        '"balance": "99999999999999999999999820 aer", "nonce": "18", ' \
        '"state": {"nonce": "18", "balance": "UrfS3MgM0uP//0w="}, ' \
        '"is_state_proof": false, ' \
        '"enc_key": "47QHSuhBTZ4b7nbw2zJz5EARyQ4XS8gDxhFDeu5mYeKNqfHhw' \
        'RutUmja3WRxV1suB12eWBeDZ"}'
    acc2 = Account.from_json(data=acc2_json_str, password='5678')
    assert acc.private_key and acc2.private_key
    assert bytes(acc.private_key) == bytes(acc2.private_key)
    assert str(acc.private_key) == str(acc2.private_key)
    assert acc.address and acc2.address
    assert bytes(acc.address) == bytes(acc2.address)
    assert str(acc.address) == str(acc2.address)
    assert acc2.nonce == 18
    assert str(acc2.balance) == "99999999999999999999999820 aer"
    assert acc2.balance.aergo == "99999999.99999999999999982 aergo"
    assert acc2.balance.gaer == "99999999999999999.99999982 gaer"
    assert acc2.balance.aer == "99999999999999999999999820 aer"

    acc3_json_str = '{' \
        '"address": "AmNW9YMMm48jTxX5Yee6tRYLpuptWdg3cqbD2CVoee1YcUGBHfad", ' \
        '"balance": "100 aergo", "nonce": "7", ' \
        '"state": {"nonce": "20", "balance": "UrfS3MgM0uP//0w=", ' \
        '"unknown_field": "1234"}, "is_state_proof": false, ' \
        '"enc_key": "47QHSuhBTZ4b7nbw2zJz5EARyQ4XS8gDxhFDeu5mYeKNqfHhwRutUmj' \
        'a3WRxV1suB12eWBeDZ"}'
    acc3 = Account.from_json(data=acc3_json_str, password='5678')
    assert acc.private_key and acc3.private_key
    assert bytes(acc.private_key) == bytes(acc3.private_key)
    assert str(acc.private_key) == str(acc3.private_key)
    assert acc.address and acc3.address
    assert bytes(acc.address) == bytes(acc3.address)
    assert str(acc.address) == str(acc3.address)
    assert acc3.nonce == 20
    assert str(acc3.balance) == "99999999999999999999999820 aer"
    assert acc3.balance.aergo == "99999999.99999999999999982 aergo"
    assert acc3.balance.gaer == "99999999999999999.99999982 gaer"
    assert acc3.balance.aer == "99999999999999999999999820 aer"

    # if is_state_proof == True, "state" will be ignored.
    acc4_json_str = '{' \
        '"address": "AmNW9YMMm48jTxX5Yee6tRYLpuptWdg3cqbD2CVoee1YcUGBHfad", ' \
        '"balance": "100 aergo", "nonce": "7", ' \
        '"state": {"nonce": "20", "balance": "UrfS3MgM0uP//0w=", ' \
        '"unknown_field": "1234"}, "is_state_proof": true, ' \
        '"enc_key": "47QHSuhBTZ4b7nbw2zJz5EARyQ4XS8gDxhFDeu5mYeKNqfHhwRutUmj' \
        'a3WRxV1suB12eWBeDZ"}'
    acc4 = Account.from_json(data=acc4_json_str, password='5678')
    assert acc.private_key and acc4.private_key
    assert acc.address and acc4.address
    assert bytes(acc.private_key) == bytes(acc4.private_key)
    assert str(acc.private_key) == str(acc4.private_key)
    assert bytes(acc.address) == bytes(acc4.address)
    assert str(acc.address) == str(acc4.address)
    assert acc4.nonce == -1
    assert str(acc4.balance) == "0 aer"
    assert acc4.balance.aergo == "0 aergo"
    assert acc4.balance.gaer == "0 gaer"
    assert acc4.balance.aer == "0 aer"



import json
from aergo.herapy.utils.converter import (
    convert_bytes_to_hex_str,
    convert_bytes_to_int_str,
)


def test_blockchain(aergo) -> None:
    print("------ Get Node Info -----------")
    node_state = aergo.get_node_state()
    print("Node State: {}".format(str(node_state)))
    node_info = aergo.get_node_info()
    print("Node Info: {}".format(str(node_info)))

    print("------ Get Blockchain Info -----------")
    blockchain_info = aergo.get_chain_info()
    print("Blockchain Info: {}".format(str(blockchain_info)))
    blockchain_info = aergo.get_chain_info(with_consensus_info=False)
    print("Blockchain Info: {}".format(str(blockchain_info)))

    print("------ Get Consensus Info -----------")
    consensus_info = aergo.get_consensus_info()
    print("Consensus Info: {}".format(str(consensus_info)))

    print("------ Get Blockchain Status -----------")
    blockchain_status = aergo.get_status()
    print("Blockchain Status: {}".format(str(blockchain_status)))
    print("Best Block Hash      = {}"
          .format(str(blockchain_status.best_block_hash)))
    print("Best Block Height    = {}"
          .format(blockchain_status.best_block_height))
    print("Best Chain ID Hash         = {}"
          .format(blockchain_status.best_chain_id_hash))
    print("base58(Best Chain ID Hash) = {}"
          .format(blockchain_status.best_chain_id_hash_b58))
    consensus_info = blockchain_status.consensus_info
    print("Consensus Info       = {}".format(consensus_info))
    print("Consensus Info: type     = {}".format(consensus_info.type))
    print("Consensus Info: status   = {}".format(consensus_info.status))
    print("Consensus Info: LIB hash = {}".format(consensus_info.lib_hash))
    print("Consensus Info: LIB no   = {}".format(consensus_info.lib_no))

    best_block_hash, best_block_height = aergo.get_blockchain_status()
    print("Best Block Hash      = {}".format(best_block_hash))
    print("str(Best Block Hash)     = {}".format(str(best_block_hash)))
    print("bytes(Best Block Hash)   = {!r}".format(bytes(best_block_hash)))
    print("int(Best Block Hash)     = {}"
          .format(convert_bytes_to_int_str(bytes(best_block_hash))))
    print("hex(Best Block Hash)     = {}"
          .format(convert_bytes_to_hex_str(bytes(best_block_hash))))
    print("Best Block Height    = {}".format(best_block_height))

    print("------ Get Blockchain Headers -----------")
    block_headers = aergo.get_block_headers(
        block_height=best_block_height, offset=best_block_height - 1)
    for i, b in enumerate(block_headers):
        print('[{}] block: {}'.format(i, str(b)))

    print("------ Get Blockchain metas -----------")
    block_metas = aergo.get_block_metas(
        block_height=best_block_height, offset=best_block_height - 1)
    for i, b in enumerate(block_metas):
        print('[{}] block: {}'
              .format(i, json.dumps(b.json(header_only=True), indent=2)))
    block_meta = aergo.get_block_meta(block_height=1)
    print("Block meta: ", block_meta)

    print("------ Get Block Status -----------")
    block = aergo.get_block(best_block_hash)
    print("Block Chain ID Hash         = ", block.chain_id_hash)
    print("base58(Block Chain ID Hash) = ", block.chain_id_hash_b58)
    print("Block Hash = ", block.hash)
    print("  Height = ", block.height)
    print("  Timestamp = ", block.timestamp)
    print("  Blocks Root Hash = ", block.blocks_root_hash)
    print("  Txs Root Hash = ", block.txs_root_hash)
    print("  Confirm = ", block.confirms)
    print("  Pub Key = ", block.public_key)
    print("  Sign = ", block.sign)
    print("  previous block hash =\n{}".format(block.prev))
    print("  Txs = ", block.tx_list)
    print(json.dumps(block.json(), indent=2))

    block = aergo.get_block(block_height=best_block_height)
    print(block)
    print("Block Hash = ", block.hash)
    print("  Height = ", block.height)
    print("  Timestamp = ", block.timestamp)
    print("  Blocks Root Hash = ", block.blocks_root_hash)
    print("  Txs Root Hash = ", block.txs_root_hash)
    print("  Confirm = ", block.confirms)
    print("  Pub Key = ", block.public_key)
    print("  Sign = ", block.sign)
    print("  previous block hash =\n{}".format(block.prev))
    print("  Txs = ", block.tx_list)

    print("------ Get Peers -----------")
    # definition of peer needs to describe.
    peers = aergo.get_peers()
    for p in peers:
        print(p)

    print("------ Get Node State -----------")
    node_state = aergo.get_node_state()
    print(node_state)
    node_state_fmt_txt = json.dumps(node_state, indent=2, sort_keys=True)
    print(node_state_fmt_txt)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: node.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='node.proto',
  package='types',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\nnode.proto\x12\x05types\"\x94\x01\n\x0bPeerAddress\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x0e\n\x06peerID\x18\x03 \x01(\x0c\x12\x1d\n\x04role\x18\x04 \x01(\x0e\x32\x0f.types.PeerRole\x12\x0f\n\x07version\x18\x05 \x01(\t\x12\x11\n\taddresses\x18\x06 \x03(\t\x12\x13\n\x0bproducerIDs\x18\x07 \x03(\x0c\"\xa9\x01\n\x10\x41gentCertificate\x12\x13\n\x0b\x63\x65rtVersion\x18\x01 \x01(\r\x12\x0c\n\x04\x42PID\x18\x02 \x01(\x0c\x12\x10\n\x08\x42PPubKey\x18\x03 \x01(\x0c\x12\x12\n\ncreateTime\x18\x04 \x01(\x03\x12\x12\n\nexpireTime\x18\x05 \x01(\x03\x12\x0f\n\x07\x61gentID\x18\x06 \x01(\x0c\x12\x14\n\x0c\x41gentAddress\x18\x07 \x03(\x0c\x12\x11\n\tsignature\x18\x08 \x01(\x0c*C\n\x08PeerRole\x12\x11\n\rLegacyVersion\x10\x00\x12\x0c\n\x08Producer\x10\x01\x12\x0b\n\x07Watcher\x10\x02\x12\t\n\x05\x41gent\x10\x03\x62\x06proto3')
)

_PEERROLE = _descriptor.EnumDescriptor(
  name='PeerRole',
  full_name='types.PeerRole',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LegacyVersion', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Producer', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Watcher', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Agent', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=344,
  serialized_end=411,
)
_sym_db.RegisterEnumDescriptor(_PEERROLE)

PeerRole = enum_type_wrapper.EnumTypeWrapper(_PEERROLE)
LegacyVersion = 0
Producer = 1
Watcher = 2
Agent = 3



_PEERADDRESS = _descriptor.Descriptor(
  name='PeerAddress',
  full_name='types.PeerAddress',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='types.PeerAddress.address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='types.PeerAddress.port', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='peerID', full_name='types.PeerAddress.peerID', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='role', full_name='types.PeerAddress.role', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='types.PeerAddress.version', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addresses', full_name='types.PeerAddress.addresses', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='producerIDs', full_name='types.PeerAddress.producerIDs', index=6,
      number=7, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=170,
)


_AGENTCERTIFICATE = _descriptor.Descriptor(
  name='AgentCertificate',
  full_name='types.AgentCertificate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='certVersion', full_name='types.AgentCertificate.certVersion', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='BPID', full_name='types.AgentCertificate.BPID', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='BPPubKey', full_name='types.AgentCertificate.BPPubKey', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='createTime', full_name='types.AgentCertificate.createTime', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expireTime', full_name='types.AgentCertificate.expireTime', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='agentID', full_name='types.AgentCertificate.agentID', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='AgentAddress', full_name='types.AgentCertificate.AgentAddress', index=6,
      number=7, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signature', full_name='types.AgentCertificate.signature', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=173,
  serialized_end=342,
)

_PEERADDRESS.fields_by_name['role'].enum_type = _PEERROLE
DESCRIPTOR.message_types_by_name['PeerAddress'] = _PEERADDRESS
DESCRIPTOR.message_types_by_name['AgentCertificate'] = _AGENTCERTIFICATE
DESCRIPTOR.enum_types_by_name['PeerRole'] = _PEERROLE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PeerAddress = _reflection.GeneratedProtocolMessageType('PeerAddress', (_message.Message,), {
  'DESCRIPTOR' : _PEERADDRESS,
  '__module__' : 'node_pb2'
  # @@protoc_insertion_point(class_scope:types.PeerAddress)
  })
_sym_db.RegisterMessage(PeerAddress)

AgentCertificate = _reflection.GeneratedProtocolMessageType('AgentCertificate', (_message.Message,), {
  'DESCRIPTOR' : _AGENTCERTIFICATE,
  '__module__' : 'node_pb2'
  # @@protoc_insertion_point(class_scope:types.AgentCertificate)
  })
_sym_db.RegisterMessage(AgentCertificate)


# @@protoc_insertion_point(module_scope)



from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.ciphers.modes import CTR
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from cryptography.hazmat.backends import default_backend
import hashlib
import os
import copy
from typing import (
    Dict,
    Union
)

from .converter import (
    privkey_to_address
)


KEYSTORE_V1: Dict = {
    "aergo_address": "",
    "ks_version": "1",
    "cipher": {
        "algorithm": "aes-128-ctr",
        "params": {
            "iv": ""
        },
        "ciphertext": ""
    },
    "kdf": {
        "algorithm": "scrypt",
        "params": {
            "dklen": 32,
            "n": 262144,
            "p": 1,
            "r": 8,
            "salt": ""
        },
        "mac": ""
    }
}


def decrypt_keystore_v1(keystore: Dict, password: str) -> bytes:
    # check version and algorithm names
    assert keystore['ks_version'] == \
        KEYSTORE_V1['ks_version'], "Invalid keystore version"

    assert keystore['cipher']['algorithm'] == \
        KEYSTORE_V1['cipher']['algorithm'], "Invalid cipher algorithm"

    assert keystore['kdf']['algorithm'] == \
        KEYSTORE_V1['kdf']['algorithm'], "Invalid kdf algorithm"

    ct = bytes.fromhex(keystore['cipher']['ciphertext'])

    # derive cipher_key (used to decrypt cipher text) from password
    backend = default_backend()
    kdf_params = keystore['kdf']['params']
    cipher_key = derive_cipher_key(
        bytes.fromhex(kdf_params['salt']),
        kdf_params['dklen'],
        kdf_params['n'],
        kdf_params['r'],
        kdf_params['p'],
        backend,
        password
    )

    # check mac to make sure the cipher key is correct
    mac = hashlib.sha256(cipher_key[16:32] + ct).digest().hex()
    assert mac == keystore['kdf']['mac'], "Failed to verify mac"

    # decrypt cipher text to get private key bytes
    iv = bytes.fromhex(keystore['cipher']['params']['iv'])
    cipher = Cipher(AES(cipher_key[:16]), CTR(iv), backend=backend)
    decryptor = cipher.decryptor()
    privkey_raw = decryptor.update(ct) + decryptor.finalize()

    # check address matches private key
    address = privkey_to_address(privkey_raw)
    assert address == keystore['aergo_address'], "Failed to verify address"

    return privkey_raw


def encrypt_to_keystore_v1(
    privkey: bytes,
    address: str,
    password: str,
    kdf_n: int = 2**18
) -> Dict:
    assert address == \
        privkey_to_address(privkey), "address doesn't match privkey"
    backend = default_backend()

    # create random cipher_key (derived from password) to encrypt
    salt = os.urandom(16)
    cipher_key = derive_cipher_key(
        salt,
        32,
        kdf_n,
        8,  # RFC 7914 recommendation
        1,  # RFC 7914 recommendation
        backend, password
    )

    # encrypt privkey
    iv = os.urandom(16)
    cipher = Cipher(AES(cipher_key[:16]), CTR(iv), backend=backend)
    encryptor = cipher.encryptor()
    ct = encryptor.update(privkey) + encryptor.finalize()

    # calculate mac
    mac = hashlib.sha256(cipher_key[16:32] + ct).digest().hex()

    # create keystore dictionary (json)
    keystore = copy.deepcopy(KEYSTORE_V1)
    keystore['cipher']['params']['iv'] = iv.hex()
    keystore['cipher']['ciphertext'] = ct.hex()
    keystore['kdf']['params']['n'] = kdf_n
    keystore['kdf']['params']['salt'] = salt.hex()
    keystore['kdf']['mac'] = mac
    keystore['aergo_address'] = address
    return keystore


def derive_cipher_key(
    salt: bytes,
    l: int,
    n: int,
    r: int,
    p: int,
    backend,
    password: str
) -> bytes:
    kdf = Scrypt(
        salt=salt,
        length=l,
        n=n,
        r=r,
        p=p,
        backend=backend
    )
    return kdf.derive(password.encode())


def encrypt_bytes(data: bytes, password: Union[str, bytes]) -> bytes:
    """
    https://cryptography.io/en/latest/hazmat/primitives/aead/
    :param data: bytes to encrypt
    :return: encrypted  data (bytes)
    """
    if isinstance(password, str):
        password = bytes(password, encoding='utf-8')

    m = hashlib.sha256()
    m.update(password)
    hash_pw = m.digest()

    m = hashlib.sha256()
    m.update(password)
    m.update(hash_pw)
    enc_key = m.digest()

    nonce = hash_pw[4:16]
    aesgcm = AESGCM(enc_key)
    return aesgcm.encrypt(nonce=nonce,
                          data=data,
                          associated_data=b'')


def decrypt_bytes(
    encrypted_bytes: bytes,
    password: Union[str, bytes]
) -> bytes:
    """
    https://cryptography.io/en/latest/hazmat/primitives/aead/
    :param encrypted_bytes: encrypted data (bytes)
    :param password: to decrypt the exported bytes
    :return: decrypted bytes
    """
    if isinstance(password, str):
        password = password.encode('utf-8')

    m = hashlib.sha256()
    m.update(password)
    hash_pw = m.digest()

    m = hashlib.sha256()
    m.update(password)
    m.update(hash_pw)
    dec_key = m.digest()

    nonce = hash_pw[4:16]
    aesgcm = AESGCM(dec_key)
    dec_value = aesgcm.decrypt(nonce=nonce,
                               data=encrypted_bytes,
                               associated_data=b'')
    return dec_value



# -*- coding: utf-8 -*-

import hashlib
from typing import (
    List
)

from google.protobuf.json_format import MessageToJson
from aergo.herapy.utils import merkle_proof as mp


class VarProofs(list):
    """ VarProof holds the inclusion/exclusion proof of a variable state
    inside a contract state trie
    """
    def __init__(self, var_proofs, storage_keys: List[bytes]) -> None:
        self.__var_proofs = var_proofs
        self.__storage_keys = storage_keys
        list.__init__(self, var_proofs[:])

    def __str__(self) -> str:
        json_string = ""
        for proof in self.__var_proofs:
            json_string += MessageToJson(proof)
        return json_string

    @property
    def var_proofs(self):
        return self.__var_proofs

    @property
    def storage_keys(self) -> List[bytes]:
        return self.__storage_keys

    def verify_var_proof(
        self,
        root: bytes,
        var_proof,
        trie_key: bytes
    ) -> bool:
        if var_proof.key != trie_key:
            return False
        value = hashlib.sha256(var_proof.value).digest()
        ap = var_proof.auditPath

        if var_proof.bitmap:
            height = var_proof.height
            bitmap = var_proof.bitmap

            if var_proof.inclusion:
                return mp.verify_inclusion_c(
                    ap, height, bitmap, root, trie_key, value)
            else:
                return mp.verify_exclusion_c(
                    root, ap, height, bitmap, trie_key, var_proof.proofKey,
                    var_proof.proofVal
                )
        else:
            if var_proof.inclusion:
                return mp.verify_inclusion(ap, root, trie_key, value)
            else:
                return mp.verify_exclusion(
                    root, ap, trie_key, var_proof.proofKey, var_proof.proofVal)

    def verify_proof(self, root: bytes) -> bool:
        """verify that the given inclusion and exclusion proofs are correct """
        if self.__var_proofs is None or 0 == self.__len__():
            return False

        for i, storage_proof in enumerate(self.__var_proofs):
            if not self.verify_var_proof(
                    root, storage_proof, self.storage_keys[i]):
                return False
        return True



# -*- coding: utf-8 -*-

import ecdsa
import hashlib

from ecdsa.util import string_to_number
from typing import (
    Union,
)

from .address import Address

from ..utils.encryption import (
    encrypt_bytes,
    decrypt_bytes
)
from ..utils.encoding import (
    encode_private_key,
    decode_private_key,
    encode_b58,
    decode_b58
)
from ..utils.signature import (
    deserialize_sig,
    serialize_sig
)
from ..errors.general_exception import (
    GeneralException
)


DEFAULT_CURVE = ecdsa.SECP256k1


class PrivateKey:
    def __init__(
        self,
        pk: Union[str, bytes, None]
    ) -> None:
        if pk is not None:
            self.__get_key(pk)
        else:
            self.__generate_new_key()

        self.__address = Address(self.__private_key.public_key)

    def __str__(self) -> str:
        privkey = encode_private_key(self.__get_private_key_bytes())
        assert privkey
        return privkey

    def __bytes__(self) -> bytes:
        return self.__get_private_key_bytes()

    def __generate_new_key(self) -> None:
        self.__signing_key = ecdsa.SigningKey.generate(curve=DEFAULT_CURVE,
                                                       hashfunc=hashlib.sha256)
        self.__signing_key.get_verifying_key()
        self.__private_key = self.__signing_key.privkey

    def __get_singing_key_instance(
        self,
        private_key: bytes
    ) -> ecdsa.SigningKey:
        if len(private_key) > 32:   # if over default size,
            # check bytes head
            if private_key[:4] != "\x08\x02\x12\x20".encode("latin-1"):
                raise ValueError("private key is incorrect")
            else:
                private_key = private_key[4:]

        d = int.from_bytes(private_key, byteorder='big')
        """
        sk = ecdsa.SigningKey.from_secret_exponent(secexp=d,
                                                   curve=ecdsa.SECP256k1,
                                                   hashfunc=hashlib.sha256)
        """
        # to avoid 'Assert' error
        sk = ecdsa.SigningKey(_error__please_use_generate=True)
        sk.curve = ecdsa.SECP256k1
        sk.default_hashfunc = hashlib.sha256
        sk.baselen = ecdsa.SECP256k1.baselen
        pubkey_point = ecdsa.SECP256k1.generator * d
        pubkey = ecdsa.ecdsa.Public_key(
            ecdsa.SECP256k1.generator, pubkey_point)
        pubkey.order = ecdsa.SECP256k1.order
        sk.verifying_key = ecdsa.VerifyingKey.from_public_point(
            pubkey_point, ecdsa.SECP256k1, hashlib.sha256)
        sk.privkey = ecdsa.ecdsa.Private_key(pubkey, d)
        sk.privkey.order = ecdsa.SECP256k1.order
        return sk

    def __get_key(
        self,
        private_key: Union[str, bytes]
    ) -> None:
        if isinstance(private_key, str):
            privkey = decode_private_key(private_key)
            assert privkey  # for mypy Optional[bytes] -> bytes
            private_key = privkey

        self.__signing_key = self.__get_singing_key_instance(private_key)
        self.__private_key = self.__signing_key.privkey

    def __get_private_key_bytes(self) -> bytes:
        d = self.__private_key.secret_multiplier
        return d.to_bytes((d.bit_length() + 7) // 8, byteorder='big')

    def sign_msg(self, msg: bytes) -> bytes:
        r, s = self.__signing_key.sign_number(string_to_number(msg))
        order = self.__private_key.public_key.generator.order()
        return serialize_sig(r, s, order)

    def verify_sign(self, msg: bytes, sign: bytes) -> bool:
        r, s = deserialize_sig(sign)
        signature = ecdsa.ecdsa.Signature(r, s)
        return self.__private_key.public_key.verifies(
            string_to_number(msg), signature)

    def __get_multiplied_point(
        self,
        address: Union[str, Address]
    ) -> ecdsa.ellipticcurve.Point:

        if isinstance(address, str):
            address = Address(address, curve=DEFAULT_CURVE)

        opponent_pubkey = address.public_key
        if not opponent_pubkey:
            raise GeneralException("Address object must have a public key")
        return opponent_pubkey.point * self.__private_key.secret_multiplier

    def asymmetric_encrypt_msg(
        self,
        address: Union[str, Address],
        msg: Union[str, bytes]
    ) -> str:
        if isinstance(msg, str):
            msg = bytes(msg, encoding='utf-8')

        point = self.__get_multiplied_point(address)
        password = point.x() + point.y()
        password = password.to_bytes((password.bit_length() + 7) // 8,
                                     byteorder='big')

        enc_msg = encode_b58(encrypt_bytes(msg, password))
        assert enc_msg  # for mypy Optional[str] -> str
        return enc_msg

    def asymmetric_decrypt_msg(
        self,
        address: Union[str, Address],
        enc_msg: Union[str, bytes]
    ) -> bytes:
        if isinstance(enc_msg, str):
            enc_bytes = decode_b58(enc_msg)
            assert enc_bytes  # for mypy Optional[bytes] -> bytes
            enc_msg = enc_bytes

        point = self.__get_multiplied_point(address)
        password = point.x() + point.y()
        password = password.to_bytes((password.bit_length() + 7) // 8,
                                     byteorder='big')

        return decrypt_bytes(enc_msg, password)

    @property
    def public_key(self) -> ecdsa.ecdsa.Public_key:
        return self.__private_key.public_key

    @property
    def address(self) -> Address:
        return self.__address

    def get_signing_key(self) -> ecdsa.keys.SigningKey:
        return self.__signing_key



# -*- coding: utf-8 -*-

"""Transaction class."""

import enum
import hashlib
import json
from typing import (
    Union,
    Optional,
    Dict
)

from . import tx_hash as txh
from ..obj import aer
from .address import Address
from ..grpc import blockchain_pb2
from ..utils.encoding import encode_signature, decode_signature, encode_payload


@enum.unique
class TxType(enum.Enum):
    NORMAL = blockchain_pb2.NORMAL
    GOVERNANCE = blockchain_pb2.GOVERNANCE
    SC_REDEPLOY = blockchain_pb2.REDEPLOY
    SC_FEE_DELEGATION = blockchain_pb2.FEEDELEGATION
    TRANSFER = blockchain_pb2.TRANSFER
    SC_CALL = blockchain_pb2.CALL
    SC_DEPLOY = blockchain_pb2.DEPLOY


class Transaction:
    """
    Transaction data structure.
    """

    def __init__(
        self,
        from_address: Union[bytes, Address, None] = None,
        to_address: Union[bytes, Address, None] = None,
        nonce: int = 0,
        amount: Union[bytes, str, int, float] = 0,
        payload: Optional[bytes] = None,
        gas_price: int = 0,
        gas_limit: int = 0,
        read_only: bool = False,
        tx_hash: Optional[bytes] = None,
        tx_sign: Union[bytes, str, None] = None,
        tx_type=TxType.TRANSFER,
        chain_id: Optional[bytes] = None,
        # typing TODO: circular import, consider removing block details from tx
        block=None,
        index_in_block: int = -1,
        is_in_mempool: bool = False
    ) -> None:
        self.__from_address = from_address
        self.__to_address = to_address
        self.__nonce = nonce
        self.__amount = aer.Aer(amount)
        self.__payload = payload

        if isinstance(gas_price, bytes):
            gas_price = int.from_bytes(gas_price, byteorder='big')
        self.__gas_price = aer.Aer(gas_price)

        if isinstance(gas_limit, bytes):
            gas_limit = int.from_bytes(gas_limit, byteorder='little')
        self.__gas_limit = gas_limit

        if isinstance(tx_type, bytes):
            tx_type = int.from_bytes(tx_type, byteorder='little')
        self.__tx_type = TxType(tx_type)

        self.__read_only = read_only
        self.__tx_hash = txh.TxHash(tx_hash)

        if isinstance(tx_sign, str):
            tx_sign = decode_signature(tx_sign)
        self.__sign = tx_sign

        self.__chain_id = chain_id

        self.__is_in_mempool = is_in_mempool
        if is_in_mempool:
            self.__block = None
            self.__index_in_block = -1
        else:
            self.__block = block
            self.__index_in_block = index_in_block

    def calculate_hash(self, including_sign: bool = True) -> bytes:
        m = hashlib.sha256()
        # nonce
        b = self.__nonce.to_bytes(8, byteorder='little')
        m.update(b)
        # from (account)
        if self.__from_address:
            m.update(bytes(self.__from_address))
        # to (recipient)
        if self.__to_address:
            m.update(bytes(self.__to_address))
        # amount
        b = bytes(self.__amount)
        m.update(b)
        # payload
        if self.__payload:
            m.update(self.__payload)
        # gas limit
        b = self.__gas_limit.to_bytes(8, byteorder='little')
        m.update(b)
        # gas price
        b = bytes(self.__gas_price)
        m.update(b)
        # type
        b = self.__tx_type.value.to_bytes(4, byteorder='little')
        m.update(b)
        # chainIdHash
        if self.__chain_id:
            m.update(self.__chain_id)
        # sign
        if including_sign and self.__sign:
            m.update(self.__sign)

        return m.digest()

    @property
    def block(self):
        return self.__block

    @property
    def index_in_block(self) -> int:
        return self.__index_in_block

    @property
    def is_in_mempool(self) -> bool:
        return self.__is_in_mempool

    @property
    def nonce(self) -> int:
        return self.__nonce

    @nonce.setter
    def nonce(self, v: int) -> None:
        if self.__read_only:
            return

        self.__nonce = v

    @property
    def from_address(self) -> Union[bytes, Address, None]:
        return self.__from_address

    @from_address.setter
    def from_address(self, v: Union[bytes, Address, None]) -> None:
        if self.__read_only:
            return

        self.__from_address = v

    @property
    def to_address(self) -> Union[bytes, Address, None]:
        return self.__to_address

    @to_address.setter
    def to_address(self, v: Union[bytes, Address, None]) -> None:
        if self.__read_only:
            return

        self.__to_address = v

    @property
    def amount(self) -> aer.Aer:
        return self.__amount

    @amount.setter
    def amount(self, v: Union[bytes, str, int, float]) -> None:
        if self.__read_only:
            return

        self.__amount = aer.Aer(v)

    @property
    def payload(self) -> Optional[bytes]:
        return self.__payload

    @payload.setter
    def payload(self, v: Optional[bytes]) -> None:
        if self.__read_only:
            return

        self.__payload = v

    @property
    def payload_str(self) -> Optional[str]:
        if self.__payload is None:
            return None
        return encode_payload(self.__payload)

    @property
    def gas_limit(self) -> int:
        return self.__gas_limit

    @gas_limit.setter
    def gas_limit(self, v: int) -> None:
        if self.__read_only:
            return

        self.__gas_limit = v

    @property
    def gas_price(self) -> aer.Aer:
        return self.__gas_price

    @gas_price.setter
    def gas_price(self, v: aer.Aer):
        if self.__read_only:
            return

        self.__gas_price = v

    @property
    def tx_type(self):
        return self.__tx_type

    @tx_type.setter
    def tx_type(self, v):
        if self.__read_only:
            return

        self.__tx_type = TxType(v)

    @property
    def chain_id(self) -> Optional[bytes]:
        return self.__chain_id

    @chain_id.setter
    def chain_id(self, v: Optional[bytes]) -> None:
        if self.__read_only:
            return

        self.__chain_id = v

    @property
    def sign(self) -> Optional[bytes]:
        return self.__sign

    @sign.setter
    def sign(self, v: Optional[bytes]) -> None:
        if self.__read_only:
            return

        self.__sign = v

    @property
    def sign_str(self) -> Optional[str]:
        if self.__sign is None:
            return None
        return encode_signature(self.__sign)

    @property
    def tx_hash(self) -> txh.TxHash:
        if self.__read_only:
            return self.__tx_hash
        return txh.TxHash(self.calculate_hash())

    def json(self, without_block: bool = False) -> Dict:
        account = None
        recipient = None
        if self.from_address is not None:
            account = str(self.from_address)
        if self.to_address is not None:
            recipient = str(self.to_address)

        tx_json = {
            "Hash": str(self.tx_hash),
            "Body": {
                "Nonce": self.nonce,
                "Account": account,
                "Recipient": recipient,
                "Amount": str(self.amount),
                "Payload": self.payload_str,
                "GasPrice": str(self.gas_price),
                "GasLimit": self.gas_limit,
                "Sign": self.sign_str,
                "Type": self.tx_type.name,
            },
            "IsInMempool": self.__is_in_mempool,
            "IndexInBlock": self.__index_in_block,
        }

        if not without_block:
            if self.__block is None:
                tx_json["Block"] = None
            else:
                tx_json["Block"] = self.__block.json(header_only=True)

        return tx_json

    def __str__(self) -> str:
        return json.dumps(self.json(), indent=2)

    def __bytes__(self) -> bytes:
        return bytes(self.tx_hash)



import aergo.herapy as herapy


def test_sc_event(aergo) -> None:
    print("------ Payload -----------")
    """

function constructor(name)
        contract.event("movie", "robocop", 1, "I'll be back")
end

function movie_event(name)
        contract.event("movie", "arnold", 2, "It's mine")
        contract.event("movie", "terminator", 3, "No, it's mine")
end

function ani_event(name)
        contract.event("a-time!!")
        contract.event("a-time!!", "pinn", "Adventure time!!", false,
                       bignum.number("999999999999999"),
                       {"princess bubble gum", 1}
        )
        contract.event("a-time!!", "Jake", "Adventure time!!", false,
                       {"whatever", 1, {true, false},
                        bignum.number("999999999999999")}
        )

        contract.event("movie", "yoda", 3, "i'm not the ani character.")
end

abi.register(movie_event, ani_event)
    """
    payload_str = "D1ZsfUyj4q8GA6AtbW81TzSTSvqaeUvcsePsaZufptRK5qSb83fyXRWdE" \
        "gGhLWs17JfRNrM4iLjGe54fkyvsyjHVgFCHqTzJWK4Yz8uQwRqzF5x1g1twZ3jPP1Xt" \
        "J95y43RD48qXGoeeoDsmmNErQJCux6d4eLZwE344bz5hM1DdWEd22yE6ogC37RLRtRZ" \
        "5FFKm3HkxaM67G7VqWfsX5t2Tmd23Myn62hkAdNWjSBaWizqyi6ENB3VD2YJGG82bVe" \
        "q9a4zMQdK58Wxnfcfqo14Vo9Y9jsY8A5QYNRi6uotRfiYt3LFrPjPBWgyQRn4qFQypS" \
        "W5vfgFZvv9VBdFPGgoMN5zeBp7VNWXvE5o8CGAf3fuFu2gCGLS4DfCtQe7VNnsCUi7U" \
        "1AdmdupqoxMHM6kCfEtk77YyiaaSTm4xUSt7R2hwzs2Xnfh8CLxMaH2cg3rsWF4fefs" \
        "5KvnjN3F1q43og6BueK66o8NssyY41kAtincbYhDC8oHaBkMTdXD6mkhb3QECwn8dh8" \
        "k9a3KTys1tdUzf99nqpK5YrAcZFDjbrmmmtcBYwZg6zdRAkQnJptvr25HzeHMfbfERH" \
        "8P1XzYmjqYynFvUNv4utuWZ7avS2xvJoZph3nSa8Xuhbdw6TFLVCDZVfZB6j8kQbJnb" \
        "1DoEEtvhPHLMP4StcgiE9grGQuduZ5iHYtS8Nk696Fq6Y9gzp19bWD3VKarq1XuueM6" \
        "FNrzCQNY22nov4mxbM8vcxp4axdjmZNWLcdttFycazorxCNbn1YBQFzGPGzstJPN72R" \
        "B9dxV17cb1TuREKTF7mVLLeahgHsY4eVnA1m5Z8U7tFKswRDTGeXv6vCt7MstVBztdd" \
        "4sRfHTXKdFCgfoQk2zwzGS8FEVgxAP3Y4Tip4nuQd8hVyTiaApFmNnkD9byxUNCc44T" \
        "r9LAnRycc239HGpZjWBwsJvmb4jsY2fuEELXfSgQf4yZhnbcPC4DYPQLDRXc8E6DdLz" \
        "ixMk7cAfUZqododk9SWzjB9FSKtzVYWiKuyg2uZoRaxDmrtZpe6pXL8F4pzt2P2HEG6" \
        "jyFn5MT9PBWhkj11vqAjWpdRgNAbTKEb5zk9ZeER9YUAT8etk2VFpAGqmGnv9CTEQr2" \
        "rfDQeNgNYzoy44mNZonhi4n4zo9h2HmrSsatWa8oxbJJYihGs7zD65TiqRfnKg43Xx5" \
        "qD"

    payload = herapy.utils.decode_address(payload_str)
    print(''.join('{:d} '.format(x) for x in payload))

    byte_code_len = int.from_bytes(payload[:4], byteorder='little')
    print("payload: byte code length = ", byte_code_len)

    print("------ Set Sender Account -----------")
    sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
    sender_account = aergo.new_account(private_key=sender_private_key)
    print("  > Sender Address: {}".format(sender_account.address))
    print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

    print("------ Deploy SC -----------")
    tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234,
                                 retry_nonce=5)
    print("  > TX: {}".format(tx.tx_hash))
    print("{}".format(str(tx)))
    assert result.status == herapy.CommitStatus.TX_OK, \
        "    > ERROR[{0}]: {1}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
    print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

    aergo.wait_tx_result(tx.tx_hash)

    print("------ Check deployment of SC -----------")
    print("  > TX: {}".format(tx.tx_hash))
    result = aergo.get_tx_result(tx.tx_hash)
    assert result.status == herapy.TxResultStatus.CREATED, \
        "  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail)

    sc_address = result.contract_address
    print("  > SC Address: {}".format(sc_address))

    print("------ Get events -----------")
    events = aergo.get_events(sc_address, "movie", with_desc=True,
                              arg_filter={0: 'arnold'})
    print("how many 'arnold' in movie? {}".format(len(events)))
    assert len(events) == 0

    events = aergo.get_events(sc_address, "movie", with_desc=True,
                              arg_filter={0: 'robocop'})
    print("how many 'robocop' in movie? {}".format(len(events)))
    assert len(events) == 1

    event_block_no = 0
    for i, e in enumerate(events):
        event_block_no = e.block_height
        print("[{}] Event: {}".format(i, str(e)))

    events = aergo.get_events(
        sc_address, "movie", end_block_no=event_block_no - 1)
    print("in history: how many movie? {}".format(len(events)))
    assert len(events) == 0

    events = aergo.get_events(
        sc_address, "movie", start_block_no=event_block_no + 1)
    print("after: how many movie? {}".format(len(events)))
    assert len(events) == 0

    print("------ Call SC -----------")
    tx, result = aergo.call_sc(sc_address, "movie_event")
    print("movie_event tx hash: {}".format(str(tx.tx_hash)))
    aergo.wait_tx_result(tx.tx_hash)

    print("------ Get events -----------")
    events = aergo.get_events(sc_address, "movie")
    print("how many 'movie'? {}".format(len(events)))
    assert len(events) == 3
    for i, e in enumerate(events):
        print("[{}] Event: {}".format(i, str(e)))
    events = aergo.get_events(sc_address, "a-time!!")
    print("how many 'a-time!!'? {}".format(len(events)))
    assert len(events) == 0

    print("------ Call SC -----------")
    tx, result = aergo.call_sc(sc_address, "ani_event")
    print("ani_event tx hash: {}".format(str(tx.tx_hash)))
    aergo.wait_tx_result(tx.tx_hash)

    print("------ Get events -----------")
    events = aergo.get_events(sc_address, "movie")
    print("how many 'movie'? {}".format(len(events)))
    assert len(events) == 4
    events = aergo.get_events(sc_address, "a-time!!")
    print("how many 'a-time!!'? {}".format(len(events)))
    assert len(events) == 3
    for i, e in enumerate(events):
        print("[{}] Event: {}".format(i, str(e)))



import aergo.herapy as herapy


def test_tx_batch(aergo) -> None:
    print("------ Set Sender Account -----------")
    sender_private_key = \
        "eHoEcHnaxpGpgzknXjuwon8VFVrLkKHC4FckGuGkQ8depiDDfyUAWC3L"
    sender_account = aergo.new_account(private_key=sender_private_key)
    print("  > Sender Address: {}".format(sender_account.address))
    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))

    print("------ Set Receiver Account -----------")
    receiver_address = "AmMPQqRJ4pd8bS9xdkw5pjExtLjaUXaYGB5kagzHu4A9ckKgnBV2"
    print("  > Receiver Address: {}".format(receiver_address))
    receiver_account = aergo.get_account(address=receiver_address)
    print("    > account state of Sender")
    print("      - balance        = {}".format(receiver_account.balance))
    print("      - nonce          = {}".format(receiver_account.nonce))
    print("      - code hash      = {}".format(receiver_account.code_hash))
    print("      - storage root   = {}".format(receiver_account.storage_root))

    print("------ Send batch Txs -----------")
    nonce = sender_account.nonce
    txs = [
        aergo.generate_tx(
            to_address=receiver_address, nonce=nonce + 1, amount='9 aer'),
        aergo.generate_tx(
            to_address=receiver_address, nonce=nonce + 2, amount='90 aer'),
        aergo.generate_tx(
            to_address=receiver_address, nonce=nonce + 3, amount='900 aer'),
        aergo.generate_tx(
            to_address=receiver_address, nonce=nonce + 4, amount='9000 aer'),
        aergo.generate_tx(
            to_address=receiver_address, nonce=nonce + 5, amount='90000 aer'),
    ]

    txs, results = aergo.batch_tx(signed_txs=txs)
    for i, tx in enumerate(txs):
        print("[{}]".format(i))
        print("  > tx: ", str(tx))
        result = results[i]
        print("  > result: ", result)
        assert result.status == herapy.CommitStatus.TX_OK, \
            "    > ERROR[{0}]: {1}".format(
                herapy.CommitStatus(result.status).name, result.detail)
        print("    > result[{0}] : {1}"
              .format(result.tx_id, result.status.name))

    aergo.wait_tx_result(txs[-1].tx_hash)

    print("------ Check TX status -----------")
    for i, tx in enumerate(txs):
        print("[{}]".format(i))
        print("  > tx: ", str(tx))
        tx_result = aergo.get_tx_result(tx.tx_hash)
        print("      result: ", tx_result)
        assert tx_result.status == herapy.TxResultStatus.SUCCESS, \
            "  > ERROR[{0}]:{1}: {2}".format(
                tx_result.contract_address, tx_result.status, tx_result.detail)

    print("------ Get Sender Account Info -----------")
    aergo.get_account()
    print("    > account state of Sender")
    print("      - balance        = {}".format(sender_account.balance))
    print("      - nonce          = {}".format(sender_account.nonce))
    print("      - code hash      = {}".format(sender_account.code_hash))
    print("      - storage root   = {}".format(sender_account.storage_root))



import pytest
import aergo.herapy as herapy


def test_account_exp_imp(aergo) -> None:

    aergo.new_account()
    new_exp_txt = aergo.export_account(password="1234")
    print("Exported txt is {}".format(new_exp_txt))

    # fail to import with wrong pwd
    with pytest.raises(herapy.errors.GeneralException):
        aergo.import_account(new_exp_txt, password='test')

    # import with correct pwd
    account = aergo.import_account(new_exp_txt, password='1234')
    print("Account private key is {}".format(account.private_key))
    print("Account address is {}".format(account.address))

    # import with exported key
    exported_txt = "47DpVeQt14U834UP6uosqA2ahpg9FPjWNM1vd1LxX1m7MXm" \
                   "AcUy4oqVmcmZgkNVh6AMhHv1CE"
    print("Exported Data is {}".format(exported_txt))
    account = aergo.import_account(exported_txt, password='1234')
    print("Account private key is {}".format(account.private_key))
    print("Account address is {}".format(account.address))

    new_exp_txt = aergo.export_account(password='1234')
    print("Exported txt is {}".format(new_exp_txt))

    assert new_exp_txt == exported_txt, \
        "Exported account should be same as imported one"



# -*- coding: utf-8 -*-

from google.protobuf.json_format import MessageToJson

from typing import (
    Optional
)

from .var_proof import VarProofs
from ..account import Account


class SCStateVar:
    """
    SCStateVar represents each variable of a calling smart contract.
    If the variable is the 'state.var' type,
    you can skip 'array_index' and 'map_key'.
    If the variable is the 'state.array' type,
    use 'array_index' with the index number.
    If the variable is the 'state.map' type,
    use 'map_key' with the key name of the map.
    """
    def __init__(
        self,
        var_name: str,
        array_index: Optional[int] = None,
        map_key: Optional[str] = None,
        empty: bool = False
    ) -> None:
        if empty:
            return

        self.var_name = var_name
        self.array_indx = array_index
        self.map_key = map_key

        if self.array_indx is not None:
            self.storage_key = '_sv_{0}-{1}'.format(
                self.var_name, self.array_indx)
        elif self.map_key is not None:
            self.storage_key = '_sv_{0}-{1}'.format(
                self.var_name, self.map_key)
        else:
            self.storage_key = '_sv_{0}'.format(self.var_name)

    def __str__(self) -> str:
        return self.storage_key

    def __bytes__(self) -> bytes:
        return self.storage_key.encode('latin-1')


class SCState:
    """ SCState holds the inclusion/exclusion proofs of a contract
    state in the global trie and of a variable's value in the
    contract trie.
    SCState is returned by aergo.query_sc_state() for easy merkle
    proof verification give a root.
    """
    def __init__(self, account: Account, var_proofs: VarProofs) -> None:
        self.__account = account
        self.__var_proofs = var_proofs

    def __str__(self) -> str:
        account_str = MessageToJson(self.__account.state_proof)
        var_str = str(self.__var_proofs)
        return account_str + var_str

    @property
    def account(self) -> Account:
        return self.__account

    @property
    def var_proofs(self) -> VarProofs:
        return self.__var_proofs

    @var_proofs.setter
    def var_proofs(self, v: VarProofs) -> None:
        self.__var_proofs = v

    def verify_proof(self, root: bytes) -> bool:
        """Verify that the given inclusion and exclusion proofs are correct """
        if not self.__account.verify_proof(root):
            return False

        sc_root = self.__account.state_proof.state.storageRoot
        return self.__var_proofs.verify_proof(sc_root)



import pytest
import aergo.herapy as herapy


@pytest.fixture(scope="session")
def aergo() -> herapy.Aergo:
    setup_ok = True
    aergo = herapy.Aergo()
    try:
        aergo.connect('localhost:7845')
    except herapy.errors.exception.CommunicationException:
        setup_ok = False
    if not setup_ok:
        pytest.fail("Integration tests require a localhost:7845 testnet. "
                    "Use `make local_testnet`")
    return aergo



from typing import Dict
import aergo.herapy as herapy
from aergo.herapy.account import Account
from aergo.herapy.utils.encryption import (
    encrypt_to_keystore_v1,
    decrypt_keystore_v1
)
from aergo.herapy.utils.converter import (
    privkey_to_address,
)


keystore: Dict = {
    "kdf": {
        "algorithm": "scrypt",
        "mac":
            "47006a8c1c17f56991ba412a4670da0c373f949ea64918531551e49fed0272ac",
        "params": {
            "dklen": 32,
            "n": 1024,
            "p": 1,
            "r": 8,
            "salt":
                "f8e0ba7b762bbff83ff7e48af13eedf08ce05fba29d247054139ca62035"
                "7215f"
        }
    },
    "aergo_address": "AmM8Bspua3d1bACSzCaLUdstjooRLy1YqZ61Kk2nP4VfGTWJzDd6",
    "ks_version": "1",
    "cipher": {
        "algorithm": "aes-128-ctr",
        "ciphertext":
            "919b663b4039a9ea88470a6e7138acebb852820100360449e7457171ee003d08",
        "params": {
            "iv": "62fc13b3065a22ffd73804e23f47f690"
        }
    }
}


def test_account_keystore_v1() -> None:
    # create new account
    account1 = Account()
    assert account1.private_key
    privkey1 = bytes(account1.private_key)
    assert account1.address
    addr1 = bytes(account1.address)
    # encrypt to keystore
    keystore = Account.encrypt_to_keystore(account1, 'password', kdf_n=2**10)
    # decrypt keystore
    account2 = Account.decrypt_from_keystore(keystore, 'password')
    assert account2.private_key
    privkey2 = bytes(account2.private_key)
    assert account2.address
    addr2 = bytes(account2.address)
    # check the decrypted account is same with original one
    assert privkey1 == privkey2
    assert addr1 == addr2


def test_decrypt_keystore_v1() -> None:
    # check decrypted address matched the keystore address
    privkey_raw = decrypt_keystore_v1(keystore, 'password')
    address = privkey_to_address(privkey_raw)
    assert address == keystore['aergo_address']


def test_encrypt_keystore_v1() -> None:
    account = Account()
    assert account.private_key
    privkey = bytes(account.private_key)
    address = str(account.address)
    new_keystore = encrypt_to_keystore_v1(
        privkey, address, 'password', kdf_n=2**10)

    # check keystore format
    assert keystore['ks_version'] == new_keystore['ks_version']
    assert keystore['cipher']['algorithm'] == \
        new_keystore['cipher']['algorithm']
    assert keystore['kdf']['algorithm'] == new_keystore['kdf']['algorithm']


def test_aergo_import_export_account() -> None:
    hera = herapy.Aergo()
    account = hera.import_account_from_keystore(
        keystore, 'password', skip_state=True)
    assert str(account.address) == keystore['aergo_address']
    new_keystore = hera.export_account_to_keystore('password', kdf_n=2**10)
    assert new_keystore['aergo_address'] == keystore['aergo_address']



import aergo.herapy as herapy


def test_account_proof(aergo) -> None:
    address = "AmPwBMCcYbqyetVmAupjtzR8GgnTAVduZAVE9SnJPzjhWEmhjSef"

    print("------ Fund a second account to pass test in testmode -------")
    sender_private_key = \
        "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
    aergo.new_account(private_key=sender_private_key)
    simple_tx, result = aergo.send_payload(to_address=address,
                                           amount=10, payload=None,
                                           retry_nonce=3)
    assert result.status == herapy.CommitStatus.TX_OK, \
        "    > ERROR[{0}]: {1}".format(result.status, result.detail)
    print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
    aergo.wait_tx_result(simple_tx.tx_hash)

    print("------ Get Account State -----------")
    best_block_hash, best_block_height = aergo.get_blockchain_status()
    block = aergo.get_block(best_block_hash)
    root = block.blocks_root_hash
    account = aergo.get_account(address=address, proof=True, root=root)
    account2 = aergo.get_account(address=address, proof=True,
                                 compressed=False, root=root)

    print("account = {}".format(account))
    print("account.balance = {}".format(account.balance.aergo))
    print("account2 = {}".format(account2))
    print("account2.balance = {}".format(account2.balance.aergo))
    assert int(account.balance) == int(account2.balance)

    print("------ Verify inclusion proof -----------")
    assert account.state_proof.inclusion is True
    assert account2.state_proof.inclusion is True
    assert account.verify_proof(root) is True, \
        "Invalid inclusion proof compressed"
    assert account2.verify_proof(root) is True, \
        "Invalid inclusion proof compressed"

    print("------ Verify Non inclusion proof -----------")
    address = "AmMejL8z3wW2doksBzzMiWM2xTb6WtZniLkLyxwqWKiLJKK8Yvqd"
    account = aergo.get_account(address=address, proof=True, root=root)
    account2 = aergo.get_account(address=address, proof=True,
                                 compressed=False, root=root)
    print(account)
    print(account2)
    assert account.state_proof.inclusion is False
    assert account2.state_proof.inclusion is False
    assert account.verify_proof(root) is True, \
        "Invalid exclusion proof compressed"
    assert account2.verify_proof(root) is True, \
        "Invalid exclusion proof compressed"
